<?php class container extends baseobject { 
	
	protected static $fields=array(
		'name'=>validator::REQUIRED,
		'containertypeid'=>array(validator::REQUIRED, validator::INTEGER),
	);
	
	protected static $helpTexts=array(
		'name'=>'A unique identifier for the container',
		'containertypeid'=>'The type of container, e.g., Unipuck',
	);

	protected static $adminSelect='SELECT SQL_CALC_FOUND_ROWS container.*,
				containercategory.name AS containercategoryname, 
				containertype.name AS containertypename,
				containertype.positions AS positions,
				project.name AS projectname
			FROM container, containertype, containercategory, project
			WHERE containertype.id=container.containertypeid 
			AND containercategory.id=containertype.containercategoryid 
			AND container.projectid=project.id';
	protected static $normalSelect='SELECT SQL_CALC_FOUND_ROWS container.*,
				containercategory.name AS containercategoryname, 
				containertype.name AS containertypename,
				containertype.positions AS positions,
				project.name AS projectname 
			FROM container, containertype, containercategory, project
			WHERE containertype.id=container.containertypeid 
			AND containercategory.id=containertype.containercategoryid 
			AND container.projectid=project.id ';
		
	public static function create($request=array()){
		$project=project::getByName(project::SHARED);
		$request['projectid']=$project['id'];
		if('dummypin'==$request['name']){
			$pinType=containertype::getByName('Pin');
			if(!$pinType){
				throw new BadRequestException('Cannot create a dummy pin. No containertype called Pin.');
			}
			$request=array(
					'name'=>'dummypin_'.microtime(true),
					'containertypeid'=>$pinType['id'],
					'projectid'=>$project['id']
			);
		} else if(!static::canCreate()){
			$containerType=containertype::getById($request['containertypeid']);
			if(!$containerType){ throw new NotFoundException('No container type with ID '.$request['containertypeid']); }
			if((boolean)($containerType['userscancreate'])===false){
				throw new ForbiddenException('Only administrators and technicians can create a '.strtolower($containerType['name']));
			}
		}
		$ret=parent::create($request,'container');
		return $ret;
	}
	
	public static function update($id,$request=array()){
	    $container=static::getById($id);
	    if(isset($request['action']) && 'unpackAll'==$request['action']){
	        //get all containercontents where $id is parent
	        //recursive-delete each one
	        containercontent::deleteContainerContentsRecursively($id);
	        return array('updated'=>$container);
	    }
	    throw new BadRequestException('Nothing to do');
	}
	
	public static function canCreate(){
		if(session::isAdmin()){ return true; }
		return true;
		$grp=usergroup::getByName(usergroup::TECHNICIANS);
		return usergroup::userisingroup($grp['id']);
		//However, creation within a given category may be prohibited, see create()
	}
	
	
	
	public static function getfullcontents($id, $request=array()){
		$request['recursive']='yes';
		$rows=static::getcontents($id,$request);
		return array(
			'total'=>count($rows),
			'rows'=>$rows
		);
	}
	
	public static function getcontents($id, $request=array()){
		if(!isset($request['recursive'])){ 
			$contents=containercontent::getByProperties(array(
					'parent'=>$id,
					'iscurrent'=>1
			), $request);
			return $contents; 
		}
		$parent=container::getById($id);
		if(!$parent){ throw new NotFoundException('Parent is not a container'); }
		return container::getContentsRecursively($parent);
	}
	public static function gethistoricalcontents($id, $request=array()){
		return containercontent::getByProperties(array(
				'parent'=>$id,
				'iscurrent'=>0
		), $request);
	}

	public static function getContentsRecursively($parent){
	    if(isset($parent['positions'])){
            $numPositions=$parent['positions'];
            $positions=array_fill(null,(int)$numPositions+1,null);
	    } else {
	        $positions=array();
	    }
		$positions[0]="dummy";
		$contents=containercontent::getByProperties(array(
				'parent'=>$parent['id'],
				'iscurrent'=>1
		), array());
		if($contents['rows']){
			foreach ($contents['rows'] as $c){
				if("container"==$c['childtype']){
					$c['childobject']['record']['childitems']=container::getContentsRecursively($c['childobject']['record']);
				} else if("crystal"==$c['childtype']){
					$shipmentId=$c['shipmentid'];
					$crystalId=$c['childobject']['record']['id'];
					$diffractionRequests=diffractionrequest::getByProperties(array(
							'crystalid'=>$crystalId,
					));
					$shippingComment='';
					$diffractionType='';
					$observedResolution='';
					$requiredResolution='';
					$minimumResolution='';
					if($diffractionRequests['rows']){
					    foreach($diffractionRequests['rows'] as $dr){
							if($dr['shipmentid']==$shipmentId){
							    $shippingComment=$dr['comment'];
							    $diffractionType=$dr['diffractiontype'];
							    $observedResolution=$dr['observedresolution'];
								$requiredResolution=$dr['requiredresolution'];
								$minimumResolution=$dr['minimumresolution'];
							}
						}
					}
					$c['childobject']['record']['shippingcomment']=$shippingComment;
					$c['childobject']['record']['diffractiontype']=$diffractionType;
					$c['childobject']['record']['observedresolution']=$observedResolution;
					$c['childobject']['record']['requiredresolution']=$requiredResolution;
					$c['childobject']['record']['minimumresolution']=$minimumResolution;
					
				}
				$c['childobject']['record']['objecttype']=$c['childobject']['objecttype'];
				$positions[$c['position']]=$c['childobject']['record'];
				$positions[$c['position']]['containercontentid']=$c['containercontentid'];
			}
		}
		return $positions;
	}
	
}

?>