
var Crystal={
		
		diffractionTypes:[
			{ label:'OSC', value:'OSC' },
			{ label:'SAD', value:'SAD' },
			{ label:'MAD', value:'MAD' },
		],
		
		/**
		 * Crystal space groups, by category.
		 * Format: 
		 * Underscores are purely to indicate digit grouping boundaries. Some numbers are written as subscripts. Where 
		 * two digits are together, the second should be rendered as a subcript, e.g. P_42_2_2 rendered to the user as 
		 * P4<sub>2</sub>22 and saved to the database as P4222.
		 */
		spaceGroups:[
			{
				'category':'Triclinic', 
				'groups':['Unknown','P_1']
			},
			{
				'category':'Monoclinic',
				'groups':[
					'P_2','P_21','C_2'
				]
			},
			{
				'category':'Orthorhombic',
				'groups':[
					'P_2_2_2','P_2_2_21','P_21_21_2','P_21_21_21','C_2_2_21','C_2_2_2','F_2_2_2','I_2_2_2','I_21_21_21'
				]
			},
			{
				'category':'Tetragonal',
				'groups':[
					'P_4','P_41','P_42','P_43','I_4','I_41','P_4_2_2','P_4_21_2','P_41_2_2',
					'P_41_21_2','P_42_2_2','P_42_21_2','P_43_2_2','P_43_21_2','I_4_2_2','I_41_2_2'
				]
			},
			{
				'category':'Trigonal',
				'groups':[
					'P_3','P_31','P_32','R_3','H_3','P_3_1_2','P_3_2_1','P_31_1_2','P_31_2_1',
					'P_32_1_2','P_32_2_1','R_3_2','H_3_2'
				]
			},
			{
				'category':'Hexagonal',
				'groups':[
					'P_6','P_61','P_65','P_62','P_64','P_63','P_6_2_2',
					'P_61_2_2','P_65_2_2','P_62_2_2','P_64_2_2','P_63_2_2'
				]
			},
			{
				'category':'Cubic',
				'groups':[
					'P_2_3','F_2_3','I_2_3','P_21_3','I_21_3','P_4_3_2','P_42_3_2','F_4_3_2',
					'F_41_3_2','I_4_3_2','P_43_3_2','P_41_3_2','I_41_3_2'
			 	]
			}
		],
		
		getThumbnailLink: function(crystal){
			var dropName=Plate.getWellNameFromObject(crystal);
			return '<a href="'+Crystal.getUrl(crystal)+'"><img src="/dropimagethumb/'+crystal.dropimageid+'/" style="max-height:5em;margin:0.25em 0" /></a>';
		},
		getTextLink: function(crystal){
			var dropName=Plate.getWellNameFromObject(crystal);
			dropName=dropName.replace("."," drop ");
			return '<a href="'+Crystal.getUrl(crystal)+'">'+dropName+' crystal '+crystal.numberindrop+'</a>';
		},
		getUrl: function(crystal){
			var dropName=Plate.getWellNameFromObject(crystal);
			return '/imagingsession/'+crystal.imagingsessionid+'/#'+dropName+'c'+crystal.numberindrop;
		},

		spaceGroupField:function(crystal,parentElement){
			var buttons='';
			var val=crystal.spacegroup;
			var displayValue="Unknown";
			if(""!=val){
				Crystal.spaceGroups.each(function(cat){
					cat.groups.each(function(grp){
						if(val==grp.replace(/_/g,"")){
							displayValue=grp.replace(/(\d)(\d)/g,"$1<sub>$2</sub>").replace(/_/g,'');
						}
					});
				});
			}
			var content='<input data-apiurl="/api/crystal/'+crystal.id+'" type="hidden" name="spacegroup" value="'+val+'" /><span class="sgvalue">'+displayValue+'</span>';
			content+=' <button onclick="Crystal.showSpaceGroupPicker(this);return false">Change...</button>';
			var ff=ui.formField({
				'label':'Space group',
				'content':content
			}, parentElement);
			ff.down("input").oldvalue=val;
			
		},
		
		showSpaceGroupPicker:function(btn){
			var box;
			if($("modalBox")){
				$("modalBox").oldTitle=$("modalBox").down("h2").innerHTML;
				$("modalBox").down("h2").innerHTML="Choose space group";
				var modalBody=$("modalBox").down(".boxbody");
				var box=modalBody.clone(false);
				box.id="sgoverlay";
				$("modalBox").down("h2").insert({ after:box });
				modalBody.style.display="none";
				box.style.display="block";
			} else {
				box=ui.modalBox({ title:"Choose a space group", content:"" });
			}
			var lbl=btn.up("label");
			lbl.down("input").oldvalue=lbl.down("input").value;
			box.field=lbl;
			Crystal.spaceGroups.each(function(cat){
				buttons='<h4>'+cat.category+'</h4>';
				cat.groups.each(function(grp){
					var btnVal=grp.replace(/_/g,"");
					var btnLbl=grp.replace(/(\d)(\d)/g,"$1<sub>$2</sub>").replace(/_/g,'');
					buttons+='<button onclick="Crystal.setSpaceGroupField(this)" style="line-height:2em; margin:0.25em" value="'+btnVal+'">'+btnLbl+'</button>';
				});
				box.innerHTML+=buttons;
			});
			
		},
		
		setSpaceGroupField:function(btn){
			var field=btn.up(".boxbody").field;
			field.down(".sgvalue").innerHTML=btn.innerHTML;
			var val=btn.value;
			if("Unknown"==val){ val=""; }
			field.down("input").value=val;
			ui.updateFormField(field.down("input"));
			if($("sgoverlay")){
				var box=$("sgoverlay").up(".box");
				$("sgoverlay").remove();
				box.down(".boxbody").style.display="block";
				if($("modalBox").oldTitle){ $("modalBox").down("h2").innerHTML=$("modalBox").oldTitle; }
			} else {
				ui.closeModalBox();
			}
		},
		
		writeDetailsFormFields(xtal,canUpdate, parentElement){
			var xtalUrl='/api/crystal/'+xtal.id;
			var diffreq;
			xtal.diffractionrequests.each(function(dr){
				if(""==dr.shipmentid){
					diffreq=dr;
				}
			});

			
			parentElement.innerHTML=parentElement.innerHTML+"";

			Crystal.spaceGroupField(xtal, parentElement);
			
			//Unit cell
			var tbl='<table style="width:100%">';
			tbl+='<tr><td id="uca">a: </td><td id="ucb">b: </td><td id="ucc">c: </td></tr>';
			tbl+='<tr><td id="ucalpha">&alpha;: </td><td id="ucbeta">&beta;: </td><td id="ucgamma">&gamma;: </td></tr>';
			tbl+='</table>';
			ui.formField({label:'Unit cell dimensions', content:tbl }, parentElement);

			parentElement.innerHTML+='<h4>Information for next synchrotron trip</h4>';

		
			//Sample name
			ui.formField({
				label:'Sample name for synchrotron',
				content:'<span id="snamectrl" style="white-space:nowrap">'+xtal.prefix+ (""==xtal.suffix ? "" : "_"+xtal.suffix) +'</span>'
			}, parentElement);
			if(canUpdate){
				$("snamectrl").innerHTML=xtal.prefix+'<span id="nameseparator">_</span>';
				ui.textBox({ apiUrl:xtalUrl, name:'suffix', value:xtal.suffix }, $("snamectrl") );
				$("suffix").observe("keyup",function(){
					if(''==$("suffix").value){
						$("nameseparator").style.visibility="hidden";
					} else {
						$("nameseparator").style.visibility="visible";
					}
				});
			}

			if(!diffreq){ return; }
			var drUrl='/api/diffractionrequest/'+diffreq.id;

			//Shipping comment
			ui.formField({
				label:'Comments',
				content:'<span id="commctrl">'+ (""==diffreq.comment ? "-" : diffreq.comment) +'</span>'
			}, parentElement);
			if(canUpdate){
				$("commctrl").innerHTML='';
				ui.textBox({ apiUrl:drUrl, name:'comment', value:diffreq.comment }, $("commctrl") );
				$("commctrl").down("input").style.width="95%";
			}

			//Diffraction type, resolutions
			ui.formField({ label:'Diffraction type', content:'<span id="dtype"></span>'}, parentElement);
			ui.formField({ label:'Observed resolution (&#8491;)', content:'<span id="obsres"></span>'}, parentElement);
			ui.formField({ label:'Required resolution: (&#8491;)', content:'<span id="reqres"></span>'}, parentElement);
			ui.formField({ label:'Minimum resolution: (&#8491;)', content:'<span id="minres"></span>'}, parentElement);
			ui.dropdownElement({ apiUrl:drUrl, name:'diffractiontype', value:diffreq.diffractiontype, options:Crystal.diffractionTypes }, $("dtype") );

			ui.textBox({ apiUrl:drUrl, name:'observedresolution', value:diffreq.observedresolution }, $("obsres") );
			ui.textBox({ apiUrl:drUrl, name:'requiredresolution', value:diffreq.requiredresolution }, $("reqres") );
			ui.textBox({ apiUrl:drUrl, name:'minimumresolution', value:diffreq.minimumresolution }, $("minres") );
			$("obsres").down("input").style.width="4em";
			$("reqres").down("input").style.width="4em";
			$("minres").down("input").style.width="4em";
			
			ui.textBox({ apiUrl:xtalUrl, name:'unitcella',     value:xtal.unitcella },     $("uca")     );
			ui.textBox({ apiUrl:xtalUrl, name:'unitcellb',     value:xtal.unitcellb },     $("ucb")     );
			ui.textBox({ apiUrl:xtalUrl, name:'unitcellc',     value:xtal.unitcellc },     $("ucc")     );
			ui.textBox({ apiUrl:xtalUrl, name:'unitcellalpha', value:xtal.unitcellalpha }, $("ucalpha") );
			ui.textBox({ apiUrl:xtalUrl, name:'unitcellbeta',  value:xtal.unitcellbeta },  $("ucbeta")  );
			ui.textBox({ apiUrl:xtalUrl, name:'unitcellgamma', value:xtal.unitcellgamma }, $("ucgamma") );
		},

		/**
		 * Deletes the crystal on the server. 
		 * @param {object} crystal The crystal to be deleted
		 * @param {function|null} onSuccess A function to be called if deletion succeeds. The response JSON will be passed as a parameter.
		 * @param {function|null} onFailure A function to be called if deletion fails.
		 */
		delete: function(crystal, afterSuccess, afterFailure){
			if(1==crystal.isFished){
				alert("Crystal cannot be deleted because it has been fished.");
				if(afterFailure){ return afterFailure(); }
				return false;
			}
			new Ajax.Request('/api/crystal/'+crystal.id, {
				method:"delete",
				postBody:"csrfToken="+csrfToken,
				onSuccess:function(transport){ 
					if(afterSuccess){
						afterSuccess(transport.responseJSON);
					}
				},
				onFailure:function(transport){ 
					AjaxUtils.checkResponse(transport);
					if(afterFailure){
						afterFailure(transport);
					}
				},
			});
		},
		
}

var PlateWidget={
		
		/**
		 * Renders the HTML for a plate widget and attaches it to the parent.
		 * @param details A JSON object containing at least the following:
		 * 					wellDrops An array of welldrop objects for this plate
		 * 					plateType A platetype object describing the plate geometry
		 *                Optional parameters:
		 *                  id The HTML ID of the widget element. Internal element IDs are prepended with this ID.
		 *                  dropPickerPosition:left Float the drop picker to the left, instead of on top.
		 *                  dropRenderer: A function taking the individual drop div, applied to all non-empty drop divs.
		 *                  dropOnMouseOver,
		 *                  dropOnMouseOut, 
		 *                  dropOnClick: Functions taking the EVENT (not the drop div element). event.target is the element.
		 *                               Applied to all non-empty drop divs.
		 * @param parent The parent HTML element
		 * @returns The plate widget element
		 */
		render:function(details, parent){
			
			var drops=details.wellDrops;
			var plateType=details.plateType;
			if(!drops || !plateType){
				alert("Cannot render plate widget - plate type and/or well drops not provided");
				return false;
			}
			//wrapper div
			var w=document.createElement("div");
			w.plateType=plateType;
			w.details=details;
			w.addClassName("platewidget");
			if(details.id){ w.id=details.id; }
			
			//drop mapping/picker
				var dp=document.createElement("table");
				dp.addClassName("pw_droppicker");
				var dm=plateType.dropmapping;
				var rows=dm.split(",");
				rows.each(function(r){
					var tr=document.createElement("tr");
					for(i=0;i<r.length;i++){
						td=document.createElement("td");
						var cellContent=r[i];
						td.innerHTML=cellContent;
						if(parseInt(cellContent)){
							td.addClassName("pw_drop pw_empty drop"+cellContent);
							td.dataset.dropnumber=cellContent;
						} else if("R"==cellContent){
							td.addClassName("pw_reservoir");
						}
						tr.appendChild(td);
					}
					dp.appendChild(tr);
				})
				w.appendChild(dp);
				if("left"==details.dropPickerPosition){ dp.style.float="left"; }
				if(plateType.subs<=1){
					dp.style.display="none";
				}

				//plate table
				var t=document.createElement("table");
				t.addClassName("pw_plate");
					//Header row
					var tr=document.createElement("tr");
					var th=document.createElement("th");
					tr.appendChild(th);
					for(c=1;c<=plateType.cols;c++){
						th=document.createElement("th");
						th.innerHTML=c;
						tr.appendChild(th);
					}
					t.appendChild(tr);

					//One row per plate type row
					for(r=1;r<=plateType.rows;r++){
						tr=document.createElement("tr");
						tr.addClassName("row"+r);
						th=document.createElement("th");
						th.innerHTML=Plate.rowLabels[r];
						tr.appendChild(th);
						for(c=1;c<=plateType.cols;c++){
							td=document.createElement("td");
							td.addClassName("pw_well pw_empty row"+r+" col"+c);
							td.dataset.row=r;
							td.dataset.col=c;
							if(details.id){ td.id=details.id+"_"+r+"_"+c; }
							for(d=1;d<=plateType.subs;d++){
								var dd=document.createElement("div");
								dd.addClassName("pw_drop pw_empty row"+r+" col"+c+" drop"+d);
								dd.dataset.row=r;
								dd.dataset.col=c;
								dd.dataset.dropnumber=d;
								if(details.id){ dd.id=details.id+"_"+r+"_"+c+"_"+d; }
//								dd.innerHTML=d
								td.appendChild(dd);
							}	
							tr.appendChild(td);
						}
						t.appendChild(tr);
					}
				w.appendChild(t);
			if("absolute"!=parent.style.position){ parent.relativize(); }
			parent.appendChild(w);

			//iterate through drops and attach, removing class name "empty"
			drops.each(function(d){
				var cell=t.down("tr.row"+d.row).down("td.col"+d.col);
				var dropDiv=cell.down(".drop"+d.dropnumber);
				dropDiv.wellDrop=d;
				dropDiv.dataset.welldropid=d.id;
				dropDiv.removeClassName("pw_empty");
				cell.removeClassName("pw_empty");
				if(dp){
					dp.down("td.drop"+d.dropnumber).removeClassName("pw_empty");
				}
			});
			
			if(dp){
				dp.select("td.pw_drop").each(function(td){
					if(!td.hasClassName("pw_empty")){
						td.observe("click",PlateWidget.dropPickerSetDrop);
					}
				});
			}
			
			w.setCurrentDrop=function(rowNumber,colNumber,dropNumber){
				return PlateWidget.setCurrentDrop(w,rowNumber,colNumber,dropNumber);
			}

			var currentRow=1;
			var currentCol=1;
			var currentDrop=1;
			if(details.currentDropIndex && details.currentDropIndex<=details.wellDrops.length){
				var wd=details.wellDrops[details.currentDropIndex];
				currentRow=wd.row;
				currentCol=wd.col;
				currentDrop=wd.dropnumber;
			}
			w.setCurrentDrop(currentRow,currentCol,currentDrop);

			window.setTimeout(function(){PlateWidget.afterRender(w)},50);
			return $(w);
		},
		/**
		 * called by PlateWidget.render. Do not call directly.
		 * @param plateWidget
		 */
		afterRender:function(plateWidget){
			var p=plateWidget.down(".pw_plate");
			p.absolutize();
			var offset=p.positionedOffset();
			var newHeight=plateWidget.getHeight()-offset.top;
			var newWidth=plateWidget.getWidth()-offset.left;
			var cellHeight=newHeight/(1+(1*plateWidget.plateType.rows));
			var cellWidth=newWidth/(1+(1*plateWidget.plateType.cols));
			cellSize=Math.min(cellHeight,cellWidth);
			var firstWell=p.down("td");
			var cellBorder=parseInt(window.getComputedStyle(firstWell).borderLeftWidth);
			cellSize-=(1*cellBorder);
			p.select("th").each(function(c){
				c.style.height=cellSize+"px";
				c.style.lineHeight=cellSize+"px";
				c.style.width=cellSize+"px";
				c.style.minWidth=cellSize+"px";
			})
			if(plateWidget.down("table.pw_droppicker")){
				plateWidget.down("table.pw_droppicker").style.marginTop=cellSize+"px";
			}
			window.setTimeout(function(){PlateWidget.attachEventsAndRenderDrops(plateWidget)},10);
		},
		/**
		 * called by PlateWidget.afterRender. Do not call directly.
		 * @param plateWidget
		 */
		attachEventsAndRenderDrops:function(plateWidget){
			var details=plateWidget.details;
			if(!details){ return; }
			plateWidget.select("div.pw_drop").each(function(d){
				if(d.hasClassName("pw_empty")){ return; /*from this iteration*/ }
				d.observe("click",PlateWidget.dropDivSetDrop);
				if(details.dropRenderer){ details.dropRenderer(d); }
				if(details.dropOnClick){ d.observe("click",details.dropOnClick); }
				if(details.dropOnMouseover){ d.observe("mouseover",details.dropOnMouseover); }
				if(details.dropOnMouseout){ d.observe("mouseout",details.dropOnMouseout); }
			});
			if(document.all /*ancient IE*/ || (!!window.MSInputMethodContext && !!document.documentMode)/*IE11*/){
				//because IE is special
				var firstTd=plateWidget.down("table.pw_plate td");
				var borderWidth=parseInt(firstTd.getStyle("borderLeftWidth"))
				var h=(firstTd.getHeight()-(2*borderWidth))+"px";
				var w=(firstTd.getWidth()-(2*borderWidth))+"px";
				plateWidget.select("div.pw_drop").each(function(d){
					d.setStyle({ width:w, height:h })
				});
			}
		},
		
		
		dropPickerSetDrop:function(evt){
			var td=evt.target;
			if(td.hasClassName("pw_current")){ return false; }
			var newDropNumber=td.dataset.dropnumber;
			var w=td.up(".platewidget");
			w.setCurrentDrop(null,null,newDropNumber);
		},
		
		dropDivSetDrop:function(evt){
			var d=evt.target;
			if(d.up(".pw_drop")){ d=d.up(".pw_drop"); }
			d.dataset.foobar="hello"
			PlateWidget.setCurrentDrop(d.up(".platewidget"),d.dataset.row, d.dataset.col, null);
		},
		
		setCurrentDrop:function(plateWidget,rowNumber,colNumber,dropNumber){
			if(null!=rowNumber){ plateWidget.currentRow=rowNumber; }
			if(null!=colNumber){ plateWidget.currentCol=colNumber; }
			if(null!=dropNumber){ 
				plateWidget.currentDrop=dropNumber; 
				if(plateWidget.down(".pw_droppicker")){
					plateWidget.down(".pw_droppicker").select("td").each(function(td){
						td.removeClassName("pw_current");
						if(td.hasClassName("drop"+dropNumber)){
							td.addClassName("pw_current");
						}
					});
					plateWidget.down(".pw_plate").select("div.pw_drop").each(function(dd){
						dd.removeClassName("pw_current");
						if(dd.dataset.dropnumber==dropNumber){
							dd.addClassName("pw_current");
						}
					});
				}
			}
			//set current on plate cell
			var plateTable=plateWidget.down(".pw_plate");
			var currentWell=plateTable.down("td.pw_current");
			if(currentWell){ currentWell.removeClassName("pw_current"); }
			currentWell=plateTable.down("tr.row"+plateWidget.currentRow).down("td.col"+plateWidget.currentCol);
			currentWell.addClassName("pw_current");
			return currentWell.down("div.drop"+plateWidget.currentDrop);
		}

};

var Plate={

		/**
		 * Renders the HTML for a plate widget and attaches it to the parent.
		 * @see PlateWidget.render for documentation. This is an alias for that.
		 */
		plateWidget:function(details, parent){
			return PlateWidget.render(details, parent);
		},
		
		//array is zero-based, plate row numbers are 1-based, so empty item first.
		rowLabels:['','A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T'],
		
		constructs:null,
		plateType:null,
		
		/*
		 * Selected coordinate ranges for setting a construct. These defaults do all drops in the entire plate.
		 * These are 1-based, ie A1 drop 1 is selectionDrop, selectionTop and selectionLeft all 1.
		 */
		selectionDrop:-1,
		selectionTop:-1,
		selectionBottom:10000,
		selectionLeft:-1,
		selectionRight:10000,

		getWellName:function(rowNumber, colNumber, dropNumber){
			var col=colNumber+"";
			if(colNumber<10){ col="0"+colNumber; }
			return Plate.rowLabels[rowNumber]+ col +"."+dropNumber;
		},
		
		getWellNameFromObject:function(obj,field){
			return Plate.getWellName(obj.row, obj.col, obj.dropnumber);
		},
		
		
		
		/*
	 	 * PLATE INSPECTIONS tab
	 	 */
		
		getManualImagers: function(){
			if(!document.manualImagers){
				new Ajax.Request('/api/imager/manualimaging/1?sortby=name&pagesize=100:pagenumber=1',{
					method:'get',
					onFailure:function(transport){
						document.manualImagers=null;
					},
					onSuccess:function(transport){
						document.manualImagers=transport.responseJSON.rows;
					},
				});
			}
		},
		
		beginCreateInspection: function(){
			if(!canCreateInspections){
				alert("Either manual creation of plate inspections is not enabled in Icebear, or you do not have permission to create them on this plate.");
				return false;
			}
			if(!document.manualImagers){
				alert("No imagers suitable for manual imaging are defined in Icebear.");
				return false;
			}
			var mb=ui.modalBox({
				id:'inspectiondetails',
				title:"New plate inspection on plate "+data.name
			});
			Plate.renderSetInspectionDetailsForm(mb);
		},
		
		beginEditInspection: function(btn){
			if(!canCreateInspections){
				alert("Either manual editing of plate inspections is not enabled in Icebear, or you do not have permission to edit them on this plate.");
				return false;
			}
			var inspection=btn.up("tr").rowData;
			if(1!=inspection.manualimaging){
				alert("Either manual editing of plate inspections is not enabled in Icebear, or this inspection was done on an automatic imager and cannot be edited.");
				return false;
			}
			Plate.editInspection(inspection);
		},
		
		editInspection:function(inspection){
			var ts=ui.modalTabSet();
			ts.imagingsession=inspection;
			var basicsTab=ts.tab({
				id:'inspectiondetails',
				label:'Basic details'
			});
			var imagesTab=ts.tab({
				id:'inspectionimages',
				label:'Images'
			});
			new Ajax.Request('/api/imagingsession/'+inspection.id+'/dropimage?pagesize=10000&pagenum=1',{
				method:'get',
				onFailure:function(transport){
					if(404==transport.status){
						imagesTab.images=[];
					} else {
						return AjaxUtils.checkResponse(transport);
					}
				},
				onSuccess:function(transport){
					imagesTab.images=transport.responseJSON.rows;
				}
			})
			Plate.renderSetInspectionDetailsForm($('inspectiondetails_body'),inspection);
			window.setTimeout(function(){
				Plate.renderSetInspectionImagesForm(inspection);
				$("inspectionimages").click();
			},50);
		},
		
		renderSetInspectionDetailsForm: function(parent,inspection){
			
			var imagerId=0;
			var imagedTime='';
			var formAction='/api/imagingsession';
			var formMethod="post";
			if(inspection){
				imagerId=inspection.imagerid;
				imagedTime=inspection.imageddatetime;
				formAction+="/"+inspection.id;
				formMethod="patch";
			}
			imagerOptions=[];
			document.manualImagers.each(function(im){
				if(im.manualimaging){
					imagerOptions.push({ label:im.name+" ("+im.temperature+"C)", value:im.id });
				}
			});

			var f=parent.form({
				method:formMethod,
				action:formAction
			});
			
			f.hiddenField('plateid',data.id);
			f.hiddenField('name','');
			f.hiddenField('manufacturerdatabaseid',0);
			
			f.datePicker({
				label:'Inspection date and time',
				name:'imageddatetime',
				value:imagedTime,
				showTime:true,
				minuteStep:5
			});
			f.dropdown({
				label:'Imaging device',
				name:'imagerid',
				value:imagerId,
				options:imagerOptions
			})
			f.formField({
				label:'Plate',
				content:data.name
			});
			
			if(!inspection){
				var btn=f.createButton({
					beforeSubmit:function(){
						f.down("input[name=name]").value=data.name+" "+new Date().getTime();
					},
					afterSuccess:function(responseJSON){
						Plate.editInspection(responseJSON.created);
						$("inspections").refresh();
					}
				});
			} else {
				f.select("label").each(function(lbl){
					lbl.afterUpdate=Plate.afterInspectionUpdate;
				});
			}
		},
		
		renderSetInspectionImagesForm: function(inspection){
			if(!inspection){
				alert('Inspection ID not provided');
				return false;
			}
			if(undefined===$("inspectionimages").images){ 
				window.setTimeout(function(){Plate.renderSetInspectionImagesForm(inspection)}, 50);
				return false;
			}
			var w=Plate.plateWidget({
				plateType:data.plateType,
				wellDrops:data.welldrops,
				dropPickerPosition:"left",
				dropRenderer:Plate.renderImageEditorDrop,
				dropOnClick:Plate.onImageEditorDropClick,
				dropOnMouseover:Plate.onImageEditorDropMouseover,
				dropOnMouseout:Plate.onImageEditorDropMouseout
			},$("inspectionimages_body"));
			if(0!=$("inspectionimages").images.length){
				$("inspectionimages").images.each(function(im){
					var div=w.down("tr.row"+im.row).down("td.col"+im.col).down("div.drop"+im.dropnumber);
					div.image=im;
				});
			}
		},

		renderImageEditorDrop:function(dropDiv){
			dropDiv.innerHTML="";
			var im=dropDiv.image;
			if(im){
				var img=document.createElement("img");
				var imageAspectRatio=im.pixelwidth/im.pixelheight;
				var containerAspectRatio=dropDiv.getWidth()/dropDiv.getHeight();
				if(imageAspectRatio>=containerAspectRatio){
					img.style.width="100%";
				} else {
					dropDiv.style.textAlign="center";
					img.style.height="100%";
				}
				img.src="/dropimagethumb/"+im.id;
				dropDiv.appendChild(img);
			}
			var overlay=document.createElement("label");
			overlay.addClassName("pw_overlay");
			overlay.setStyle({
				position:"absolute",
				top:0, left:0, right:0, bottom:0,
				backgroundSize:"50% 50%", backgroundPosition:"center center", backgroundRepeat:"no-repeat",
				textAlign:"center",
				display:"block",
				opacity:0.0
			});
			dropDiv.appendChild(overlay);
			if(im){
				overlay.style.backgroundImage='url(/skins/'+skin+'/images/icons/btn_remove.gif)';
				overlay.title="Remove this image";
			} else {
				overlay.style.backgroundImage='url(/skins/'+skin+'/images/icons/btn_add.gif)';
				overlay.title="Add an image";
				var f=document.createElement("input");
				f.type="file";
				f.observe("change",Plate.startDropImageAdd);
				overlay.appendChild(f);
				overlay.htmlFor=f;
				f.setStyle({ position:"absolute",top:0,left:0,height:"100%",width:"100%", opacity:"0.0" });
			}
		},		

		onImageEditorDropClick:function(evt){
			var dropDiv=evt.target;
			if(!dropDiv.hasClassName(".pw_drop")){ dropDiv=dropDiv.up(".pw_drop"); }
			if(dropDiv.image){
				Plate.startDropImageRemove(dropDiv);
			} else {
				//Handled by file input and its onchange event
			}
			return true;
		},
		onImageEditorDropMouseover:function(evt){
			var dropDiv=evt.target;
			if(dropDiv.tagName.toLowerCase()=="input"){ dropDiv=dropDiv.up(".pw_drop"); }
			if(!dropDiv.hasClassName("pw_overlay")){ dropDiv=dropDiv.down(".pw_overlay"); }
			dropDiv.style.opacity=0.6;
			if(dropDiv.image){
				//show delete icon
			} else {
				//show add icon
			}
		},
		onImageEditorDropMouseout:function(evt){
			var dropDiv=evt.target;
			if(dropDiv.tagName.toLowerCase()=="input"){ dropDiv=dropDiv.up(".pw_drop"); }
			if(!dropDiv.hasClassName("pw_overlay")){ dropDiv=dropDiv.down(".pw_overlay"); }
			dropDiv.style.opacity=0;
		},
		
		startDropImageAdd:function(evt){
			var fileField=evt.target; //the file input
			fileField.disabled="disabled";
			var dropDiv=fileField.up(".pw_drop");
			dropDiv.addClassName("updating");
			dropDiv.down("label").style.background="none"; //Don't show (+) on mouseover
			var formData=new FormData();
			formData.append('dropimage', fileField.files[0], fileField.files[0].name)
			formData.append('csrfToken', encodeURIComponent(csrfToken));
			formData.append('imagingsessionid', encodeURIComponent($("modalTabSet").imagingsession.id));
			formData.append('welldropid', encodeURIComponent(dropDiv.dataset.welldropid));
			var xhr=new XMLHttpRequest();
			var method="post";
	        xhr.open(method, "/api/dropimage", true);
	        xhr.onload=function(){
	        	try{
	        		xhr.responseJSON=JSON.parse(xhr.responseText);
	        	} catch(ex) {
	        		alert("Could not understand reply from server:\n\n"+xhr.responseText);
	        		return false;
	        	}
	            if(200==xhr.status || 201==xhr.status){
	                if(xhr.responseJSON.created){
	                	dropDiv.image=xhr.responseJSON.created;
	                }
	            } else if(401==xhr.status){
	            	ui.handleSessionExpired();
	            } else if(xhr.responseJSON.error){
	            	alert("Image upload failed. The server said:\n\n"+xhr.responseJSON.error);
	            } else {
	            	alert("Image upload failed. The server said:\n\n"+xhr.responseText);
	            }
	            dropDiv.removeClassName("updating");
	            Plate.renderImageEditorDrop(dropDiv);
	        }			
			xhr.send(formData);
			
		},
		
		startDropImageRemove:function(dropDiv){
			if(!dropDiv.image){ alert("No image attached to that drop"); return false; }
			var dropName=Plate.rowLabels[dropDiv.dataset.row];
			if(10>(1*dropDiv.dataset.col)){ dropName+="0"; }
			dropName+=dropDiv.dataset.col+" drop "+dropDiv.dataset.dropnumber;
			if(!confirm("Really remove image from "+dropName+"?")){
				return false;
			}
			dropDiv.down("label").style.background="none"; //Don't show (x) on mouseover
			dropDiv.addClassName("updating");
			new Ajax.Request('/api/dropimage/'+dropDiv.image.id,{
				method:'delete',
				postBody:'csrfToken='+csrfToken,
				onSuccess:function(transport){ Plate.dropImageRemove_onSuccess(transport,dropDiv); },
				onFailure:function(transport){ Plate.dropImageRemove_onFailure(transport,dropDiv); },
			});
		},
		dropImageRemove_onSuccess:function(transport,dropDiv){
			AjaxUtils.checkResponse(transport);
			if(transport.responseJSON.deleted){ dropDiv.image=null; }
            Plate.renderImageEditorDrop(dropDiv);
			dropDiv.removeClassName("updating");
		},
		dropImageRemove_onFailure:function(transport,dropDiv){
			AjaxUtils.checkResponse(transport);
            Plate.renderImageEditorDrop(dropDiv);
			dropDiv.removeClassName("updating");
		},
		
		afterInspectionUpdate:function(){
			$("inspections").refresh();
		},
		
		/*
	 	 * PROTEIN tab
	 	 */

	 	/**
	 	 * Begins the process of setting a protein.
	 	 * @param handler A function to clean up the page after setting
	 	 */
		beginSetProtein: function(handler){
			if("Default Project"==data.projectname){
				new Ajax.Request('/api/project/?pagesize=1000',{
					method:'get',
					onSuccess:function(transport){
						if(!AjaxUtils.checkResponse(transport)){ return false; }
						Plate.showProjectList(transport.responseJSON, handler);
					},
					onFailure:function(transport){
						AjaxUtils.checkResponse(transport);
					},
				});
			} else {
				Plate.showProjectList({'rows':[{ name:data.projectname, id:data.projectid, description:'Plate is in project "'+data.projectname+'". You can choose from proteins in that project' }]}, handler);
			}
		},
		
		showProjectList: function(projects, handler){
			var mb=ui.modalBox({
				'title':'Choose the construct',
				'content':''
			});
			Plate.doAfter=handler;
			projects.rows.each(function(p){
				if(1==p.isarchived || 1==p.issystem){ return; } //from this iteration
				var ti=mb.treeItem({
					record:p,
					id:"project"+p.id,
					header:p.name+": "+p.description+"",
					updater:Plate.showProteinsForTarget
				});
			});
			if(1==mb.select(".treeitem").length){
				//Plate.showProteinsForTarget(mb.down(".treeitem"));
				mb.select(".treehead")[0].click();
			}		
		},

		showProteinsForTarget: function(elem){
			if(elem.up(".treeitem")){ elem=elem.up(".treeitem"); }
			var proj=elem.record;
			new Ajax.Request('/api/project/'+proj.id+'/protein?pagesize=1000',{
				method:'get',
				onSuccess:function(transport){
					if(!AjaxUtils.checkResponse(transport)){ return false; }
					elem.down(".treebody").innerHTML='';
					transport.responseJSON.rows.each(function(p){
						var ti=elem.treeItem({
							record:p,
							id:"protein"+p.id,
							header:p.name+" ("+p.proteinacronym+"): "+p.description+"",
							updater:Plate.showConstructsForProtein
						});
					});
				},
				onFailure:function(transport){
					if(404==transport.status){
						elem.down(".treebody").innerHTML='<label style="text-align:left">No proteins in this project</label>';
						return;
					}
					AjaxUtils.checkResponse(transport);
				},
			});
		},

		showConstructsForProtein: function(elem){
			if(elem.up(".treeitem")){ elem=elem.up(".treeitem"); }
			var prot=elem.record;
			new Ajax.Request('/api/protein/'+prot.id+'/construct?pagesize=1000',{
				method:'get',
				onSuccess:function(transport){
					if(!AjaxUtils.checkResponse(transport)){ return false; }
					elem.down(".treebody").innerHTML='';
					transport.responseJSON.rows.each(function(c){
						var lbl=document.createElement("label");
						lbl.className="treeitem"
						lbl.record=c;
						lbl.innerHTML='<input type="button" value="Use this construct" onclick="Plate.setConstruct(this)" style="cursor:pointer" /> '+c.name+": "+c.description;
						lbl.style.textAlign="left";
						lbl.style.padding="0.5em 0.25em";
						elem.down(".treebody").appendChild(lbl);
					});
				},
				onFailure:function(transport){
					if(404==transport.status){
						elem.down(".treebody").innerHTML="No constructs for this protein";
						return;
					}
					AjaxUtils.checkResponse(transport);
				},
			});
		},

		setConstruct: function(btn){
			var construct=btn.up(".treeitem").record;
			/*
			on success, do modalBox.doAfter and close mb
			on failure close mb
			*/
			new Ajax.Request("/api/plate/"+data.id, {
				method:'patch',
				parameters:{
					csrfToken:csrfToken,
					constructid:construct.id,
					selectionDrop:Plate.selectionDrop,
					selectionTop:Plate.selectionTop,
					selectionBottom:Plate.selectionBottom,
					selectionLeft:Plate.selectionLeft,
					selectionRight:Plate.selectionRight,
				},
				onSuccess:function(transport){ 
					AjaxUtils.checkResponse(transport);
					Plate.doAfter(transport);
					ui.closeModalBox();
					$("files").refresh();
				},
				onFailure:function(transport){ 
					AjaxUtils.checkResponse(transport);
				},
			});
		},

		
		afterSetConstruct: function(transport){
			data=transport.responseJSON.updated;
			Plate.constructs=null;
			Plate.renderProteinTab();
			Plate.renderScreenTab();
			var projectLink=$("project_label").down("a");
			projectLink.href="/project/"+data.projectid;
			projectLink.innerHTML=data.projectname;
			ui.closeModalBox();
		},

		/**
		 * Renders the "Protein" tab.
		 * TODO If more than one construct per plate is allowed, UI.
		 */
		renderProteinTab:function(){
			$("proteins_body").innerHTML='';
			if(null==Plate.constructs){
				window.setTimeout(Plate.getConstructs,50);
				return;
			}
			var out='';
			var f=$("proteins_body").form({
				id:'proteinstabform',
				action:'/api/plate/'+data.id,
				method:'patch'
			});
			if(0==Plate.constructs.length){
				f.textField({
					label:'No protein is set',
					name:'noprotein',
					value:'',
					readonly:true
				});
				if(canEdit){
					f.buttonField({
						label:'Choose the protein...',
						onclick:function(){ Plate.beginSetProtein(Plate.afterSetConstruct);return false; },
						readonly:true
					});
				}
				out='<a href="#" onclick="Plate.beginSetProtein(Plate.afterSetConstruct);return false">Set protein construct...</a>';
			} else if(1==Plate.constructs.length){
				var c=Plate.constructs[0];
				f.textField({
					label:'Protein',
					name:'proteinname',
					value:c.proteinname,
					readonly:true
				});
				f.textField({
					label:'Protein acronym',
					name:'proteinacronym',
					value:c.proteinacronym,
					readonly:true
				});
				f.textField({
					label:'Construct',
					name:'constructname',
					value:c.name+': '+c.description,
					readonly:true
				});
				if(canEdit){
					f.buttonField({
						label:'Change the protein...',
						onclick:function(){ Plate.beginSetProtein(Plate.afterSetConstruct);return false; },
						readonly:true
					});
				}
				window.setTimeout(Plate.showProteinConcentration,50);
				
			} else {
				//TODO multi construct plate ui, and
				//plate.showProteinSolution();
			}
			
		},

		getWellDrops:function(){
			new Ajax.Request("/api/plate/"+data.id+"/welldrop",{
				method:'get',
				onSuccess:Plate.getWellDrops_onSuccess,
				onFailure:Plate.getWellDrops_onFailure,
			});
		},
		getWellDrops_onSuccess:function(transport){
			if(!AjaxUtils.checkResponse(transport)){ return false; }
			data.welldrops=transport.responseJSON.rows;
		},
		getWellDrops_onFailure:function(transport){ 
			return AjaxUtils.checkResponse(transport);
		},

		showProteinConcentration:function(){
			if(!data.welldrops){
				window.setTimeout(Plate.showProteinConcentration,100);
				return false;
			}
			/*
			 * Initial implementation assumes this is the same across all wells/drops
			 */
			var wd=data.welldrops[0];
			var proteinBuffer=wd.proteinbuffer
			var concAmount=wd.proteinconcentrationamount;
			var concUnit=wd.proteinconcentrationunit;
			var f=$("proteinstabform");
			f.textArea({
				label:'Protein buffer',
				name:'proteinbuffer',
				value:proteinBuffer,
				readonly:!canEdit
			});
			var concContent=concAmount+concUnit.replace("u","&#181;");
			if(canEdit){
				concContent='<input type="text" id="proteinconcentrationamount" name="proteinconcentrationamount" value="'+concAmount+'" style="width:3em"/>';
				concContent+='<select id="proteinconcentrationunit" name="proteinconcentrationunit">';
				concContent+='<option value="mg/mL" '+( 'mg/mL'==concUnit ?  'selected="selected"' : '' )+'>mg/mL</option>';
				concContent+='<option value="ug/uL" '+( 'ug/uL'==concUnit ?  'selected="selected"' : '' )+'>&#181;g/&#181;L</option>';
				concContent+='</select>';
			}
			f.formField({
				label:'Protein concentration',
				content:concContent
			});
			if(canEdit){
				fieldValidations["proteinconcentrationamount"]=["required","float"];
				$("proteinbuffer").observe("keyup",Plate.updateProteinConcentration);
				$("proteinconcentrationamount").observe("keyup",Plate.updateProteinConcentration);
				$("proteinconcentrationunit").observe("keyup",Plate.updateProteinConcentration);
				$("proteinconcentrationunit").observe("change",Plate.updateProteinConcentration);
			}
		},
		
		updateProteinConcentration:function(evt){
			var field=evt.target;
			window.clearTimeout(submitTimer);
			if(!validator.validate(field)){ return false; }
			if(field.up("label,td")){ field.up("label,td").removeClassName("invalidfield"); }
			window.submitTimer=setTimeout(function(){ Plate.doUpdateProteinConcentration(field) },1500);
		},
		doUpdateProteinConcentration:function(field){
			field.up("label").addClassName("updating");
			new Ajax.Request('/api/plate/'+data.id, {
				method:'patch',
				postBody:"csrfToken="+csrfToken+"&"+field.name+"="+field.value, //Defaults on server to all drops in all wells
				onSuccess:Plate.updateProteinConcentration_onSuccess,
				onFailure:Plate.updateProteinConcentration_onFailure
			});
		},
		updateProteinConcentration_onSuccess:function(transport){
			if(!AjaxUtils.checkResponse(transport)){ return false; }
			$("proteinbuffer").up("label").removeClassName("updating");
			$("proteinconcentrationamount").up("label").removeClassName("updating");
			data.wellDrops=transport.responseJSON.rows;
		},
		updateProteinConcentration_onFailure:function(transport){
			return AjaxUtils.checkResponse(transport);
		},
		
		getConstructs: function(){
			new Ajax.Request("/api/plate/"+data.id+'/construct', {
				method:'get',
				onSuccess:function(transport){ 
					AjaxUtils.checkResponse(transport);
					Plate.constructs=transport.responseJSON.rows;
					Plate.renderProteinTab();
					return;
				},
				onFailure:function(transport){ 
					if(404==transport.status){
						Plate.constructs=new Array();
						Plate.renderProteinTab();
						return;
					}
					Ajax.checkResponse(transport); 
				},
			});
		},
		
	  	/*
	  	 * SCREEN tab
	  	 */
		renderScreenTab: function(){
			var tab=$("screens_body");
			var pageSize=Math.max(96, (data.cols*data.rows));
			tab.dataset.apiurl='/api/screen/'+data.screenid+'/screencondition?pagenum=1&pagesize='+pageSize;
			if('Default Project'==data.projectname){
				if(canEdit){
					tab.innerHTML='Set the protein first';
				} else {
					tab.innerHTML='No screen has been chosen for this plate.';
				}
				return false;
			}
			if(""==data.screenid){
				tab.innerHTML='';
				if(canEdit){
					var ss=document.createElement("div");
					ss.setStyle({ cssFloat:"left", width:"49%" });
					var sf=tab.form({ action:'/api/screen/', method:'post', id:'standardscreenform' });
					sf.hiddenField("name","Optimization "+data.name);
					sf.hiddenField("projectid", data.projectid);
					sf.hiddenField("rows", "");
					sf.hiddenField("cols", "");
					sf.formField({ label:'<h3>Choose a standard screen</h3>',content:'&nbsp;' });
					sf.formField({ label:'',content:'' });
					sf.formField({ label:'Standard screen',content:'<input type="button" value="Choose..." onclick="Plate.beginSetScreen(Plate.afterSetScreen)"/>' });
					ss.appendChild(sf);
					tab.appendChild(ss);

					var os=document.createElement("div");
					var of=tab.form({ action:'/api/screen/', method:'post', id:'optimizationscreenform' });
					of.setStyle({ cssFloat:"right", width:"49%", marginBottom:"1.5em" });
					of.formField({ label:'<h3>Upload an optimization screen file<h3>',content:'&nbsp;' });
					of.formField({ label:'',content:'<input type="file" name="file" onchange="Plate.uploadScreen()" id="createscreenbtn"/>' });
					of.hiddenField("name", "Optimization "+data.name);
					of.hiddenField("projectid", data.projectid);
					of.hiddenField("rows", "");
					of.hiddenField("cols", "");
					os.appendChild(of);
					tab.appendChild(os);

					tab.innerHTML+='<div class="tabhelp">'+
					'<p style="clear:both">You can either (i) use a standard screen or (ii) upload details of an optimization screen. In both cases, the screen will be attached to your plate as a file.</p>'+
					'<h3 style="">Standard screens:</h3>'+
					'<p>If your plate uses a standard screen, simply click the <strong>Choose...</strong> button above and choose your screen from the list.</p>'+
					'<p>To add more standard screens to Icebear, talk to your administrator.</p>'+
					'<h3 style="clear:both;margin-top:1.5em">Optimization screens:</h3>'+
					'<p>You can upload a file describing your optimization screen. Select your file in "Upload an optimization screen file" above, and it will be uploaded and attached to your plate automatically.</p>'+
					'<p>Icebear understands the following screen file formats:</p>'+
					'<ul style="padding-left:1.5em">'+
					'<li><strong>Two-column CSV:</strong> A simple file with well number and description. You can use this <a href="/resources/WellAndCondition.csv">Template CSV file</a>;</li>'+
					'<li><strong>Mimer:</strong> The output from Mimer*, saved as CSV (not Excel format);</li>'+
					'<li><strong>Tecan CrysScreen:</strong> The Microsoft Word document generated by the Tecan robot can be uploaded directly into Icebear; or</li>'+
					'<li><strong>Rigaku CrystalTrak:</strong> A plate or screen XML file generated by CrystalTrak. If Icebear is configured to import from your Rigaku imagers, these files should be detected automatically; you shouldn\'t need to upload them here.</li>'+
					'</ul>'+
					'<p>You can upload other files, but Icebear won\'t understand them. Instead, the file will be attached and each condition will read "See screen definition file."</p>'+
					'<hr style="clear:both;margin-top:1.5em;margin-bottom:1.5em"/>'+
					'<p><span style="font-size:80%">* Mimer: an automated spreadsheet-based crystallization screening system. Brodersen DE, Andersen GR, Andersen CB<br/>'+
						'<span style="color:transparent">* </span>Acta Cryst F, 2013 (69) pp815-820. <a target="_blank" href="https://www.ncbi.nlm.nih.gov/pubmed/?term=PMID%3A+23832216">PubMed PMID:23832216</a></span></p>'+
					'</div>';
				}
			} else {
				new Ajax.Request('/api/screen/'+data.screenid, {
					method:'get',
					onSuccess:function(transport){
						if(!AjaxUtils.checkResponse(transport)){ return false; }
						new Ajax.Request(tab.dataset.apiurl,{
							method:'get',
							onSuccess:function(transport){
								if(!AjaxUtils.checkResponse(transport)){ return false; }
								var screen=transport.responseJSON.screen;
								var contentBefore='<div style="text-align:right">'+
									'<span style="float:left"><a href="/screen/'+screen.id+'">'+screen.name+'</a>';
									if(1==screen.isstandard){
										contentBefore+=' (Standard screen)';
									} else {
										contentBefore+=' (Optimization screen)';
									}
									contentBefore+='</span>';
									if(canEdit){
										contentBefore+='<input type="button" value="Change..." onclick="Plate.beginSetScreen(Plate.afterSetScreen)"';
									}
									contentBefore+='</div>';
								var headers=['Well number','Description'];
								var cellTemplates=['{{wellnumber}}','{{description}}'];
								$("screens_body").table({
									'headers':headers,
									'cellTemplates':cellTemplates,
									'contentBefore':contentBefore
								}, transport.responseJSON);
							},
							onFailure:AjaxUtils.checkResponse
						});
					}, 
					onFailure:AjaxUtils.checkResponse
				});			
			}
		},
	  	 
		beginSetScreen: function(handler){
			new Ajax.Request('/api/screen/isstandard/1?pagesize=25',{
				method:'get',
				onSuccess:function(transport){
					if(!AjaxUtils.checkResponse(transport)){ return false; }
					Plate.showScreenList(transport.responseJSON, handler);
				},
				onFailure:function(transport){
					AjaxUtils.checkResponse(transport);
				},
			});

		},
		showScreenList: function(response, handler){
			var mb=ui.modalBox({
				'title':'Choose the screen',
				'content':'Just a moment...'
			});
			Plate.doAfter=handler;
			mb.dataset.pagesize=25;
			mb.table({
				'headers':['Name','Manufacturer',''],
				'cellTemplates':['{{name}}','{{manufacturer}}','<input type="button" value="Choose" onclick="Plate.chooseStandardScreen(this)">']
			}, response);		
		},
		chooseStandardScreen: function(btn){
			btn.up("tr").addClassName("updating");
			var screen=btn.up("tr").rowData;
			Plate.setScreen(screen.id);
		},
		uploadScreen: function(){
			var frm=$("optimizationscreenform");
			frm.rows.value=data.rows;
			frm.cols.value=data.cols;
			if(0==frm.file.files.length){ alert('Choose a screen description file to upload'); return false; }
			frm.file.options={ afterSuccess:Plate.setUploadedScreen };
			ui.submitForm(frm.file);
		},
		setUploadedScreen: function(responseJSON){
			var screen=responseJSON.created;
			Plate.setScreen(screen.id);
		},
		
		setScreen: function(screenId){
			new Ajax.Request('/api/plate/'+data.id,{
				method:'patch',
				parameters:{
					csrfToken:csrfToken,
					screenid:screenId
				},
				onSuccess:function(transport){
					if(!AjaxUtils.checkResponse(transport)){ return false; }
					Plate.afterSetScreen(transport);
				},
				onFailure:function(transport){
					AjaxUtils.checkResponse(transport);
				},
			});
		},
		afterSetScreen: function(transport){
			data=transport.responseJSON.updated;
			Plate.renderScreenTab();
			$("files_body").refresh();
			ui.closeModalBox();
		},

	 	/*
	 	 * DROP CONDITIONS tab
	 	 */
		renderDropConditionsTab:function(){
			/*
			 * First implementation, assumes same across entire plate
			 */
			if(!data.plateType || !data.welldrops){
				$("drops_body").innerHTML="Just a moment...";
				window.setTimeout(Plate.renderDropConditionsTab,100);
				return;
			}
			var units=['uL','nL'];
			var unitsOptions='';
			units.each(function(u){
				unitsOptions+='<option value="'+u+'">'+u.replace("u","&#181;")+'</option>';
			});
			
			$("drops_body").innerHTML="";
			var dropMap=PlateType.getDropMapElement(data.plateType, $("drops_body"));
			dropMap.style.height="100%";
			dropMap.style.width=dropMap.getHeight()+"px";
			var drops=dropMap.select(".dropmapwell");
			drops.each(function(drop){
				drop.style.textAlign="center";
				var dropNumber=drop.dataset.dropnumber;
				var protAmount="";
				var protUnit="uL";
				var wellAmount="";
				var wellUnit="uL";

				for(i=0;i<data.welldrops.length;i++){
					var wd=data.welldrops[i];
					if(wd.dropnumber==dropNumber){
						protAmount=wd.proteinsolutionamount;
						protUnit=wd.proteinsolutionunit;
						wellAmount=wd.wellsolutionamount;
						wellUnit=wd.wellsolutionunit;
						break;
					}
				}
				
				if(canEdit){
					var out="<h3>Well solution:</h3>";
					out+='<div style="display:inline-block;width:6.25em;border:1px solid #666;border-radius:0.25em">';
					out+='<input type="text" style="width:2em;height:1.5em" name="wellsolutionamount" value="'+wellAmount+'" />';
					out+='<select style="border:none;width:3em;height:1.8em" name="wellsolutionunit">'+unitsOptions.replace('"'+wellUnit+'"','"'+wellUnit+'" selected="selected"')+'</select>';
					out+='</div>';
					out+='<hr style="margin:0.5em"/>';
					out+='<h3>Protein solution:</h3>';
					out+='<div style="display:inline-block;width:6.25em;border:1px solid #666;border-radius:0.25em">';
					out+='<input type="text" style="width:2em;height:1.5em" name="proteinsolutionamount" value="'+protAmount+'" />';
					out+='<select style="border:none;width:3em;height:1.8em" name="proteinsolutionunit">'+unitsOptions.replace('"'+protUnit+'"','"'+protUnit+'" selected="selected"')+'</select>';
					out+='</div>';
					drop.innerHTML+=out;
					drop.select("input,select").each(function(elem){
						elem.observe("keyup",Plate.updateWellSolutions);
						elem.observe("change",Plate.updateWellSolutions);
						elem.observe("blur",Plate.updateWellSolutions);
					});
				} else {
					drop.innerHTML+='<h3>Protein solution:</h3>';
					var protContent=protAmount+protUnit.replace("u","&#181;");
					if(""==protAmount || ""==protUnit){ protContent="(Not set)"; }
					drop.innerHTML+=protContent;
					drop.innerHTML+='<hr style="margin:0.5em"/>';
					drop.innerHTML+="<h3>Well solution:</h3>";
					var wellContent=wellAmount+wellUnit.replace("u","&#181;");
					if(""==wellAmount || ""==wellUnit){ wellContent="(Not set)"; }
					drop.innerHTML+=wellContent;
				}
			});
		},

		updateWellSolutions:function(evt){
			window.clearTimeout(submitTimer);
			var field=evt.target;
			var drop=field.up("td");
			if(field.up("label,td")){ field.up("label,td").removeClassName("invalidfield"); }
			drop.select("input,select").each(function(elem){
				if(!validator.validate(elem)){
					drop.addClassName("invalidfield");
					return false;
				}
			});
			window.submitTimer=setTimeout(function(){ Plate.doUpdateWellSolutions(field) },1500);
		},
		doUpdateWellSolutions:function(field){
			/*
			 * First implementation, assumes solutions are same for all drops in same position for whole plate
			 */
			var drop=field.up("td");
			if(drop.hasClassName("updating")){
				return false;
			}
			drop.removeClassName("invalidfield");
			drop.addClassName("updating");
			var parameters={ csrfToken:csrfToken};
			parameters[field.name]=field.value;
			parameters['selectionDrop']=drop.dataset.dropnumber;
			new Ajax.Request("/api/plate/"+data.id,{
				method:"patch",
				parameters:parameters,
				onSuccess:Plate.updateWellSolutions_onSuccess,
				onFailure:Plate.updateWellSolutions_onFailure
			});
		},
		updateWellSolutions_onSuccess:function(transport){
			/*
			 * First implementation, assumes solutions are same for all drops in same position for whole plate
			 */
			if(!AjaxUtils.checkResponse(transport)){ return false; }
			var updatedDrops=transport.responseJSON.updated.updateddrops;
			if(!updatedDrops || 0==updatedDrops.length){ return false; }
			firstUpdatedDrop=updatedDrops[0];
			var dropCells=$("drops_body").select(".dropmapwell");
			dropCells.each(function(dropCell){
				if(dropCell.dataset.dropnumber==firstUpdatedDrop.dropnumber){
					['proteinsolutionamount','proteinsolutionunit','wellsolutionamount','wellsolutionunit'].each(function(field){
						dropCell.down('[name='+field+']').value=firstUpdatedDrop[field];
					});
					dropCell.removeClassName("updating");
				}
			});
		},
		updateWellSolutions_onFailure:function(transport){
			AjaxUtils.checkResponse(transport);
		},
		
		getPlateType:function(){
			new Ajax.Request('/api/platetype/'+data.platetypeid,{
				method:"get",
				onSuccess:Plate.getPlateType_onSuccess,
				onFailure:Plate.getPlateType_onFailure,
			});
		},
		getPlateType_onSuccess:function(transport){
			if(!AjaxUtils.checkResponse(transport)){ return false; }
			data.plateType=transport.responseJSON;
		},
		getPlateType_onFailure:function(transport){
			AjaxUtils.checkResponse(transport);
		},

		
		/*
		 * INSPECTIONS tab of plate view
		 */
		writeInspectionLink: function(obj,field){
			return ImagingSession.getLink(obj,field);
		},
		
		/*
		 * Best score in, e.g., Project's plates tab
		 */
		renderPlateBestScoreCell:function(item){
			if(""==item.bestscorelabel){ return '-'; }
			return '<span style="float:right;width:20px;background-color:#'+item.bestscorecolor+'">&nbsp;</span>'+item.bestscorelabel;
		},
		
}



var PlateType={
		
		getDropMapElement:function(plateType, parent){
			var t=document.createElement("table");
			t.className="dropmap";
			var dropMap=plateType.dropmapping.split(",");
			dropMap.each(function(row){
				var tr=document.createElement("tr");
				for(c=0;c<row.length;c++){
					var td=document.createElement("td");
					var content=row[c];
					if(parseInt(content)){ 
						td.className="dropmapwell";
						td.dataset.dropnumber=content;
						var lbl=document.createElement("div");
						lbl.className="welllabel";
						lbl.innerHTML=content;
						td.appendChild(lbl);
					} else if("R"==content){
						//Reservoir
						td.className="dropmapreservoir";
					} else if("E"==content){
						//Empty well
						td.className="dropmapempty";
					} else if("X"==content){
						//Dead space on the plate
						td.className="dropmapdead";
					}
					tr.appendChild(td);
				}
				t.appendChild(tr);
			});
			if(null!=parent){ parent.appendChild(t); }
			return t;
		}
		
}


var ImagingSession={
	
		getLink: function(obj,field){
			var lighttype=obj.lighttype;
			var label=obj[field];
			if('imageddatetime'==field){ label=ui.friendlyDate(label); }
			return ImagingSession.getLightPathIcon(lighttype)+'&nbsp;<a href="/imagingsession/{{id}}">'+label+'</a>';
		},
	
		getLightPathIcon: function(lightTypeName){
			var types=['Visible','UV'];
			var offset=60*Math.max(0, types.indexOf(lightTypeName));
			return '<span class="lightbulb" style="background-position:0 -'+offset+'px" title="'+lightTypeName+'">&nbsp;</span>';
		}
			
}


var CrystallizationScreen={ //don't call it Screen - almost guaranteed to conflict with some UI thing or other

		getConditionCell:function(obj,field){
			if(1==data.isstandard && canEdit){
				return '<form action="/api/screencondition/'+obj.id+'/" method="patch"><input type="text" name="description" value="'+ obj["description"] +'" /></form>';
			}
			return obj[field];
		},
		
		adjustConditionsTableForEdit:function(){
			if(!$("conditions_body_table")){
				window.setTimeout(CrystallizationScreen.adjustConditionsTableForEdit,50);
				return;
			}
			if($("conditions_body_table").adjustedForEdit){
				return;
			}
			var cells=$$('#conditions_body_table td+td');
			cells.each(function(c){
				c.style.width="85%";
				var inp=c.down("input");
				if(inp){
					inp.style.width="99%";
					inp.observe("keyup", function(){ ui.updateFormField(inp)});
					inp.dataset.oldvalue=inp.value;
				}
			});
			$("conditions_body_table").adjustedForEdit=true;
		},

}

var Container={
		
	addChildByName:function(childName,parentId,position,shipmentId,successCallback){
		Container.getByName(childName, function(found){
			Container.addChildById(found.id,parentId,position,shipmentId,successCallback);
		});
	},
	addChildById:function(childId,parentId,position,shipmentId,successCallback){
		var parameters={
			'csrfToken':csrfToken,
			'parent':parentId,
			'child':childId,
			'position':position,
		};
		if(null!=shipmentId){ parameters['shipmentid']=shipmentId; }
		new Ajax.Request('/api/containercontent',{
			method:'post',
			parameters:parameters,
			onSuccess:function(transport){
				if(!transport.responseJSON || !transport.responseJSON.created){
					return AjaxUtils.checkResponse(transport);
				}
				if(successCallback){
					successCallback(transport.responseJSON.created);
				}
			},
			onFailure:function(transport){ AjaxUtils.checkResponse(transport); },
		});
	},
		
	/**
	 * Looks for a container with name/barcode "name". If found, sets the found object as
	 * the value of the supplied placeholder. If not found, errors and sets placeholder to "Not found".
	 */
	getByName:function(name,successCallback){
		new Ajax.Request('/api/container/name/'+name, {
			method:'get',
			onSuccess:function(transport){
				var found=transport.responseJSON;
				if(found.rows){ found=found.rows[0]; }
				if(successCallback){
					successCallback(found);
				}
			},
			onFailure:function(transport){
				if(404!=transport.responseStatus){ return AjaxUtils.checkResponse; }
				alert("No container with name "+name);
			},
		});
	},
		
	getAllContents:function(container,successCallback){
		new Ajax.Request('/api/container/'+container.id+'/content?recursive=yes',{
			method:'get',
			onSuccess:function(transport){
				var found=transport.responseJSON;
				container.childitems=found;
				if(successCallback){
					successCallback(found);
				}
			},
			onFailure:function(transport){
				if(404!=transport.responseStatus){ return AjaxUtils.checkResponse;}
				alert("Container is empty");
			},
		});
	},

	showAllContents:function(container,elem, showEmptyingControls){
		var category=container.containercategoryname.toLowerCase();
		if(!elem){ 
			elem=ui.modalBox({ title:"Contents of "+category+" "+container.name }); 
		}
		if(!container.childitems){
			alert("Cannot show container contents - not set")
			return false;
		}
		if("dewar"==category){
			Container.renderDewarContents(container, elem, showEmptyingControls);
		} else if("puck"==category){
			Container.renderPuckContents(container, elem, showEmptyingControls);
		} else if("pin"==category){
			Container.renderPinContents(container, elem, showEmptyingControls);
		}
	},
	
	renderDewar:function(dewar,parentElement, showEmptyingControls){
		var	elem=ui.treeItem({
			header:"Dewar: "+dewar.name
		});
		Container.renderDewarContents(dewar,elem.down(".treebody, .boxbody, .tabbody"), showEmptyingControls);
	},
	renderDewarContents:function(dewar,parentElement, showEmptyingControls){
		var childitems=dewar.childitems;
		var isEmpty=true;
		for(var i=1;i<childitems.length;i++){
			if(childitems[i]){
				isEmpty=false;
				Container.renderPuck(childitems[i],parentElement, showEmptyingControls);
			}
		}
		if(isEmpty){
			parentElement.innerHTML="Dewar is empty.";
		} else {
			var buttons=parentElement.select(".removechild");
			if(buttons){
				buttons.each(function(btn){
					btn.stopObserving("click");
					btn.observe("click",function(evt){
						Container.removeFromParent(evt.target);
						evt.stopPropagation();
					});
				});
			}
		}
	},

	renderPuck:function(puck,parentElement, showEmptyingControls){
		if(!puck.containercategoryname || "puck"!=puck.containercategoryname.toLowerCase()){
			parentElement.treeItem({ header: puck.name+" is not a puck" });
			return false;
		}
		var controls='';
		if(showEmptyingControls){
			controls='Remove puck and <input value="Keep all contents" type="button" class="removechild" data-recursive="0" />&nbsp;<input value="Wash all pins" type="button" class="removechild" data-recursive="1" />';
		}
		var	elem=parentElement.treeItem({
			record:puck,
			header:"Puck: "+puck.name+'<span style="float:right;line-height:2.5em">'+controls+'</span>'
		}).down(".treebody");
		Container.renderPuckContents(puck,elem, showEmptyingControls);
	},
	renderPuckContents:function(puck,parentElement, showEmptyingControls){
		var childitems=puck.childitems;
		var numchilditems=childitems.length;
		for(var i=1;i<numchilditems;i++){
			if(!childitems[i]){
				childitems[i]={
					position:i,
					name:'',
					rendername:'(empty)',
					samplename:''
				}
			}
			var pin=childitems[i];
			if(0===pin.name.indexOf("dummypin")){
				pin.rendername="";
				pin.controls='Remove from puck and <input style="float:none" value="Wash pin" type="button" class="removechild" data-recursive="1" />';
			} else {
				pin.rendername=pin.name;
				pin.controls='Remove from puck and <input style="float:none" value="Keep crystal" type="button" class="removechild" data-recursive="0" />&nbsp;<input style="float:none" value="Wash pin" type="button"  class="removechild" data-recursive="1"/>';
			}
			pin.position=i;
			pin.isEmpty=0;
			if(childitems[i].childitems && childitems[i].childitems[1]){
				var xtal=childitems[i].childitems[1];
				if(xtal.name){
					pin.proteinacronym=xtal.proteinacronym ? xtal.proteinacronym : '';
					pin.samplename=xtal.name;
					pin.shippingcomment=xtal.shippingcomment? xtal.shippingcomment : '';
				} else {
					//User can't see contents
					pin.proteinacronym='(hidden)';
					pin.samplename='(hidden)';
					pin.shippingcomment='(hidden)';
				}
			} else {
				//position is empty
				pin.proteinacronym='';
				pin.samplename='';
				pin.shippingcomment='';
				pin.controls=''; //nothing to remove
				pin.isEmpty=1;
			}
		}
		var headers=['Slot','Pin','Sample name','Protein acronym','Remarks'];
		var cellTemplates=['{{position}}','{{rendername}}','{{samplename}}','{{proteinacronym}}','{{shippingcomment}}'];
		if(showEmptyingControls){
			headers=['Slot','Pin','Sample name','Protein acronym','&nbsp;'];
			cellTemplates=['{{position}}','{{rendername}}','{{samplename}}','{{proteinacronym}}','{{controls}}'];
		}
		var tbl=ui.table({
			headers:headers,
			cellTemplates:cellTemplates
		}, {rows:childitems/*.splice(1)*/}, parentElement);
		if(parentElement.down(".listmore")){
			parentElement.down(".listmore").remove();
		}
		var buttons=parentElement.select(".removechild");
		if(buttons){
			buttons.each(function(btn){
				btn.stopObserving("click");
				btn.up("td").style.textAlign="right";
				btn.observe("click",function(evt){
					Container.removeFromParent(evt.target);
					evt.stopPropagation();
				});
			});
		}
	},

	/**
	 * Use this only for a loose pin. renderPuckContents will handle pins in a puck.
	 * @param pin An object representing the pin
	 * @param parentElement The element to render into
	 * @param showEmptyingControls if true, render a Remove button
	 */
	renderPinContents:function(pin,parentElement, showEmptyingControls){
		var childitems=pin.childitems;
		var out='';
		if(2==childitems.length && !childitems[1]['name']){
			out="Pin is empty.";
		}
		for(var i=1;i<childitems.length;i++){
			if(childitems[i]){
				var xtal=childitems[i];
				var xtalName=xtal["name"] || "(hidden)";
				var construct=xtal["constructname"] || "(hidden)";
				var protein="(hidden)";
				if(xtal["proteinname"] && xtal["proteinacronym"]){
					protein=xtal["proteinname"]+" ("+xtal["proteinacronym"]+")";
				}
				out+="<h3>Crystal "+i+"</h3>";
				out+="Sample name: "+xtalName;
				out+="<br/>Protein: "+protein;
				out+="<br/>Construct: "+construct;
				if(showEmptyingControls){
					//Assumes only one crystal per pin. 
					out+='<label><input value="Wash pin" type="button" class="removechild" data-recursive="1" data-containercontentid="'+xtal.containercontentid+'"/></label>';
				}
			}
		}
		parentElement.innerHTML=out;
		var buttons=parentElement.select(".removechild");
		if(buttons){
			buttons.each(function(btn){
				btn.stopObserving("click");
				btn.up("label").style.textAlign="left";
				btn.observe("click",function(evt){
					Container.removeFromParent(evt.target);
					evt.stopPropagation();
				});
			});
		}
	},

	
	removeFromParent:function(removeButton){
		var childItemId=0;
		var recursive=1*removeButton.dataset.recursive;
		if(undefined!==removeButton.dataset.containercontentid){
			childItemId=removeButton.dataset.containercontentid;
			if(removeButton.up("label")){ removeButton.up("label").addClassName("updating"); }
		} else if(removeButton.up("tr.datarow")){ 
			tr=removeButton.up("tr.datarow");
			childItemId=tr.rowData.containercontentid;
			tr.addClassName("updating");
		} else if(removeButton.up("div.treeitem")){
			var ti=removeButton.up("div.treeitem");
			childItemId=ti.record.containercontentid;
			ti.down("h3").addClassName("updating");
		}
		if(!childItemId){ return false; }
		var uri='/api/containercontent/'+childItemId;
		var postBody="csrfToken="+csrfToken;
		if(recursive){
			postBody+="&recursive=1";
		}
		new Ajax.Request(uri,{
			method:"delete",
			postBody:postBody,
			onSuccess:function(transport){ Container.removeFromParent_onSuccess(removeButton, transport); },
			onFailure:function(transport){ Container.removeFromParent_onFailure(removeButton, transport); },
		})
	},
	removeFromParent_onSuccess:function(removeButton,transport){
		removeButton.up(".updating").removeClassName("updating");
		if(removeButton.up("tr")){
			removeButton.up("tr").select("td+td").each(function(td){
				td.innerHTML="";
			});
		} else if(removeButton.up(".treeitem")){
			removeButton.up(".treeitem").remove();
		} else if(removeButton.up(".pincontent")){
			removeButton.up(".pincontent").remove();
			ui.closeModalBox();
		}
	},
	removeFromParent_onFailure:function(removeButton,transport){
		removeButton.up(".updating").removeClassName("updating");
		AjaxUtils.checkResponse(transport);
	},
	
	
}

var Protein = {
		
		codonAsAmino:{
			"ATT":"I", "ATC":"I", "ATA":"I", "CTT":"L", "CTC":"L", "CTA":"L", "CTG":"L", "TTA":"L",
			"TTG":"L", "GTT":"V", "GTC":"V", "GTA":"V", "GTG":"V", "TTT":"F", "TTC":"F", "ATG":"M",
			"TGT":"C", "TGC":"C", "GCT":"A", "GCC":"A", "GCA":"A", "GCG":"A", "GGT":"G", "GGC":"G",
			"GGA":"G", "GGG":"G", "CCT":"P", "CCC":"P", "CCA":"P", "CCG":"P", "ACT":"T", "ACC":"T",
			"ACA":"T", "ACG":"T", "TCT":"S", "TCC":"S", "TCA":"S", "TCG":"S", "AGT":"S", "AGC":"S",
			"TAT":"Y", "TAC":"Y", "TGG":"W", "CAA":"Q", "CAG":"Q", "AAT":"N", "AAC":"N", "CAT":"H",
			"CAC":"H", "GAA":"E", "GAG":"E", "GAT":"D", "GAC":"D", "AAA":"K", "AAG":"K", "CGT":"R",
			"CGC":"R", "CGA":"R", "CGG":"R", "AGA":"R", "AGG":"R", "TAA":"*", "TAG":"*", "TGA":"*",
		},
		
		dnaToProtein:function(dna){
			dna=dna.replace(/\s/g,"");
			prot="";
			var dnaParts=dna.match(/.{1,3}/g);
			dnaParts.each(function(codon){
				var amino=Protein.codonAsAmino[codon];
				if(!amino){ return false; }
				prot+=amino;
			});
			return prot;
		},
		
		formatDnaSequence:function(seq){
			if(!seq || ""==seq){ return false; }
			var seqParts=seq.match(/.{1,3}/g);
			seq=seqParts.join(" ").toUpperCase();
			return seq;
		},

		formatProteinSequence:function(seq){
			if(!seq || ""==seq){ return false; }
			var seqParts=seq.match(/.{1,10}/g);
			seq=seqParts.join(" ").toUpperCase();
			return seq;
		},

}

var Shipment={
		
		cleanOldDewarTabs:function(currentContainers){
			$$(".containertab").each(function(head){
				var dewarName=head.dataset.containername;
				var inShipment=false;
				currentContainers.each(function(cont){
					if(cont.name==dewarName){
						inShipment=true;
					}
				});
				if(!inShipment){
					head.next(".tabbody").remove();
					head.remove();
				}
			});
			if(!$("shiptabs").down("h2.current")){
				var t=$("shipmentdetails");
				t.addClassName("current");
				window.document.location.hash=t.id;
			}
		},
		
		getAndRenderContainers:function(){
			new Ajax.Request('/api/shipment/'+data.id+'/container',{
				method:'get',
				onSuccess:function(transport){
					if(transport.responseJSON.rows){
						//First remove any tabs for dewars that are NOT in the shipment
						Shipment.cleanOldDewarTabs(transport.responseJSON.rows);
						//Then render all dewars that ARE in the shipment, reusing any existing tabs
						transport.responseJSON.rows.each(function(cont){
							if(!cont.containercategoryname){
								Shipment.renderPlate(cont);
							} else if('dewar'==cont.containercategoryname.toLowerCase()){
								Shipment.renderDewar(cont);
							} else {
								alert(cont.name)+" is directly inside the shipment but is not a plate or dewar.";
							}
						});
					}
				},
				onFailure:function(transport){
					if(404!==transport.status){
						return AjaxUtils.checkResponse(transport);
					}
					$$(".containertab").each(function(head){
						head.next(".tabbody").remove();
						head.remove();
					});
					Shipment.cleanOldDewarTabs();
				},
			});
		},
		renderDewar:function(dewar){
			var t=$(dewar.name);
			if(!t){
				t=$("shiptabs").tab({
					label:dewar.name,
					id:dewar.name,
					classes:'containertab',
					content:''
				});
			}
			t.dataset.containercontentid=dewar.containercontentid;
			t.dataset.containername=dewar.name;
			t.dataset.containerid=dewar.id;
			var tb=t.next();
			tb.innerHTML="";
			Container.renderDewarContents(dewar, tb);
			if(""!=data.dateshipped){
				tb.insert({ top:ui.infoMessageBar("This is the data sent to the synchrotron. Some information may have changed since then.")});
			} else if(userCanShip){
				//TODO Remove Dewar control
				tb.insert({ top:ui.infoMessageBar('<input type="button" class="noprint" title="Remove from shipment (keeping all contents)" style="cursor:pointer;float:none;margin-right:5em;" onclick="Shipment.removeContainer(this)" value="Remove dewar from shipment" /> Add a puck: <form class="noprint" style="display:inline" onsubmit="return Shipment.addPuckByBarcode(this)"><input class="noprint" type="text" name="barcode" style="width:12em" value="" placeholder="Scan barcode"/></form>')});
				tb.down(".msgbar").addClassName("noprint");
				tb.select(".treeitem").each(function(ti){
					//Remove Puck control
					ti.down(".treehead").insert({ bottom:'<input type="button" class="noprint" title="Remove from dewar (keeping all contents)" style="cursor:pointer;float:none;margin-left:5em;" onclick="Shipment.removeContainer(this)" value="Remove puck" /> '});
				});
				tb.select("tr.datarow").each(function(tr){
					var barcodeCell=tr.down("td").next();
					if(tr.rowData.containercontentid){
						//Insert "Remove Pin" control before barcode (or, if no barcode, loose)
						barcodeCell.insert({ 'top':'<input type="button" class="noprint" title="Remove from puck (keeping crystal in pin)" style="cursor:pointer;float:none" onclick="Shipment.removeContainer(this)" value="X" /> ' });
					} else {
						//Nothing here. Add pin by scanning barcode.
						barcodeCell.innerHTML='<form class="noprint" style="display:inline" onsubmit="return Shipment.addPinByBarcode(this)"><input class="noprint" style="width:8em" type="text" name="barcode" value="" placeholder="Scan barcode"/></form>';
					}
				});
			}
			
		},
		renderPlate:function(plate){
			
		},

		addDewarByBarcode:function(){
			var b=$("addtoplevelbybarcode");
			var barcode=b.value.trim();
			if(""==barcode){ return false; }
			var alreadyIn=false;
			var numContainers=0;
			$$('.tab').each(function(t){
				if(t.dataset.containername){
					numContainers++;
					if(barcode==t.dataset.containername){
						alreadyIn=true;
					}
				}
			});
			if(alreadyIn){
				b.value="";
				alert(barcode+" is already in this shipment");
				return false;
			}
			var tb=b.up(".tabbody");
			var bar=b.up("label");
			bar.addClassName("updating");
			var shipmentId=data.id;
			var position=numContainers+1;
			Container.getByName(barcode, function(found){
				if("dewar"!=found.containercategoryname.toLowerCase()){ 
					alert(barcode+" is not a dewar.");
					bar.removeClassName("updating");
					return false;
				}
				Container.addChildById(found.id,shipmentId,position,shipmentId,Shipment.getAndRenderContainers);
				bar.removeClassName("updating");
			})
			b.value="";
			return false; //stop form submission and page reload
		},
		
		addPuckByBarcode:function(frm){
			var barcode=frm.down("input").value.trim();
			if(""==barcode){ 
				alert("Puck barcode cannot be empty");  
				return false; 
			}
			var tb=frm.up(".tabbody");
			var bar=frm.up(".infobar");
			bar.addClassName("updating");
			var dewarId=tb.previous(".tab").dataset.containerid;
			var shipmentId=data.id;
			var position=$$(".treeitem").length+1;
			Container.getByName(barcode, function(found){
				if("puck"!=found.containercategoryname.toLowerCase()){ 
					alert(barcode+" is not a puck.");
					bar.removeClassName("updating");
					return false;
				}
				Container.addChildById(found.id,dewarId,position,null,Shipment.getAndRenderContainers);
			})
			frm.down("input").value="";
			return false; //stop form submission and page reload
		},
		
		addPinByBarcode:function(frm){
			var field=frm.down("input");
			var barcode=field.value.trim();
			if(""==barcode){ 
				alert("Pin barcode cannot be empty");  
				return false; 
			}
			var tr=frm.up("tr");
			var puckId=frm.up(".treeitem").record.id;
			var shipmentId=data.id;
			var position=tr.rowData.position;
			tr.addClassName("updating");
			Container.getByName(barcode, function(found){
				if("pin"!=found.containercategoryname.toLowerCase()){ 
					alert(barcode+" is not a pin.");
					tr.removeClassName("updating");
					return false;
				}
				//Get the contents; will 404 if no crystal in pin.
				new Ajax.Request('/api/container/'+found.id+'/content',{
					method:'get',
					onSuccess:function(transport){
						if(!transport.responseJSON){
							return AjaxUtils.checkResponse(transport);
						}
						Container.addChildById(found.id,puckId,position,null,Shipment.getAndRenderContainers);
					},
					onFailure:function(transport){
						if(404!=transport.status){
							return AjaxUtils.checkResponse(transport);
						}
						alert(barcode+" has no crystal. Cannot add it to the shipment.");
						tr.removeClassName("updating");
						field.value="";
					}
				});
			})
			return false; //stop form submission and page reload
		},
		
		removeContainer:function(btn){
			var deletedTopLevel=false;
			var containerContentId=null;
			if(btn.up("tr")){
				//Pin
				containerContentId=btn.up("tr").rowData.containercontentid;
				btn.up("tr").addClassName("updating");				
			} else if(btn.up(".treeitem")){
				//Puck
				containerContentId=btn.up(".treeitem").record.containercontentid;
				btn.up(".treeitem").addClassName("updating");
			} else if(btn.up(".tabbody")){
				//Dewar
				//TODO Could also be a plate
				containerContentId=btn.up(".tabbody").previous("h2").dataset.containercontentid;
				btn.up("div.msgbar").addClassName("updating");
				deletedTopLevel=true;
			}
			if(!containerContentId){
				alert("Cannot determine container type for removal.");
				return false;
			}
			new Ajax.Request('/api/containercontent/'+containerContentId,{
				method:'delete',
				postBody:'csrfToken='+csrfToken,
				onSuccess:function(transport){
					if(!transport.responseJSON){ return AjaxUtils.checkResponse(transport); }
					Shipment.getAndRenderContainers();
					if(deletedTopLevel){
						window.setTimeout(Shipment.renumberTopLevelContainers,50);
					}
				},
				onFailure:AjaxUtils.checkResponse
			});
			
		},
		
		/**
		 * After deleting a top level container, iterate through tabs and set the "position" number
		 * to consecutive values starting from 1. Otherwise, adding another may fail due to clashing
		 * position numbers. Strictly speaking, we don't care what position a dewar occupies within a
		 * shipment - but the server does.
		 */
		renumberTopLevelContainers:function(){
			var pos=1;
			$$(".containertab").each(function(head){
				new Ajax.Request('/api/containercontent/'+head.dataset.containercontentid,{
					method:'patch',
					parameters:{
						'csrfToken':csrfToken,
						'position':pos
					},
				});
				pos++;
			});
		},
		
		 validate:function(){
			var errors=[];
			var dewars=$$(".containertab");
			if(0==dewars.length){ errors.push("Shipment must contain at least one dewar."); }
			dewars.each(function(d){
				var dewarName=d.dataset.containername;
				var pucks=d.next().select(".treeitem");
				if(0==pucks.length){ errors.push("Dewar "+dewarName+" is empty. Remove it or add a puck."); }
				pucks.each(function(p){
					var puckName=p.record.name;
					var slots=p.select("tr.datarow");
					var numPins=0;
					slots.each(function(tr){
						if(!tr.rowData.childitems || 2<tr.rowData.childitems.length){
							//position is empty
						} else {
							numPins++;
							if(1!=tr.rowData.childitems[1].hasacronym){
								errors.push("Crystal in position "+tr.rowData.position+" of puck "+puckName+" in dewar "+dewarName+" has no protein acronym. Set the parent plate's protein.");
							}
						}
					});
					if(0==numPins){errors.push("Puck "+puckName+" in dewar "+dewarName+" is empty. Remove it or add a pin."); }
				});
			});
			
			if(0==errors.length){ return true; }
			alert("You must fix these errors before the shipment can be sent:\n\n* "+errors.join("\n* ") );
			return false;
		},
		
		
		/**
		 * Submits the shipment contents to the synchrotron
		 */
		send:function(){
			if(""!=data.dateshipped){ alert("Shipment has already shipped"); return false; }
			if(!Shipment.validate()){ return false; }
			if(!confirm("This will send the shipment data to the synchrotron.\n\nAre you sure?")){ return false; }
			new Ajax.Request('/api/shipment/'+data.id,{
				method:'patch',
				parameters:{
					csrfToken:csrfToken,
					dateshipped:new Date().toISOString().slice(0,10) //GMT date
				},
				onSuccess:Shipment.send_onSuccess,
				onFailure:Shipment.send_onFailure,
				
			});
		},
		
		send_onSuccess:function(transport){
			//data['dateshipped']=new Date().toISOString().slice(0,10); //GMT date
			//Re-render to hide certain controls, etc.- easier just to reload the page?
			var msg='Shipment is now sent.';
			if(transport.responseJSON.msg){
				msg+='\n\n'+msg;
			}
			msg+='\n\nWhen you click OK, the page will reload.';
			alert(msg);
			window.location.reload(true);
		},
		send_onFailure:function(transport){
			//TODO Could be login required at xray source
			alert("Shipment submission failed");
		},
		
		
}
