var Fishing={
		
	//True if the notes box is being shown
	showingNotes:false,
	
	//Holds an array of item IDs created during this drag-drop, for undo
	currentFish:{},

	//Holds a list of all containers on the bench, including nested containers.
	//Used to prevent scanning of containers that Icebear thinks are nested.
	containersOnBench:[],
	
	//Holds the user's configuration of the fishing UI. Values below are defaults.
	config:{
        'autopintopuck':1,
        'shownotesboxafterfish':0
	},

	/***************************************************************************
	 * 
	 * INITIALIZATION
	 * 
	 ***************************************************************************/

	/**
	 * Initialises the UI.
	 */
	init:function(){
		Fishing.loadConfig();
		$("barcodeform").onsubmit=function(){ Fishing.handleScan(); return false; }
		$("configbutton").observe("click",Fishing.toggleConfig);
		window.setInterval(Fishing.focusBarcodeField,250);
	},

	/**
	 * Returns focus to the barcode field, unless a drag operation is in progress or
	 * the notes box is being shown.
	 */
	focusBarcodeField:function(){
		if(Fishing.showingNotes){ return false; }
		$("barcode").focus();
	},

	/***************************************************************************
	 * 
	 * FISHING UI CONFIGURATION
	 * 
	 ***************************************************************************/

	/**
	 * Writes the current configuration to UserConfig.
	 */
	saveConfig:function(){
		Object.keys(Fishing.config).each(function(item){
			UserConfig.set("fishing_"+item, 1*Fishing.config[item]);
			if($(item) && $(item).up("label")){
				$(item).up("label").removeClassName("updating");
			}
		});
	},
	
	/**
	 * Loads the previously-saved user configuration.
	 */
	loadConfig:function(){
		Object.keys(Fishing.config).each(function(item){
			var val=UserConfig.get("fishing_"+item, Fishing.config[item]);
			Fishing.config[item]=1*val;
		});
	},
		
	/**
	 * Shows the configuration options in a modal box.
	 */
	toggleConfig:function(){
		var bb=ui.modalBox({ 
			title:'Configure fishing',
			content:''
		});	
		var f=bb.form({ autosubmit:false });
		f.checkbox({
			name:'autopintopuck',
			label:'After fishing a crystal onto a pin, move the pin to first empty puck position',
			value:Fishing.config.autopintopuck
		})
		f.checkbox({
			name:'shownotesboxafterfish',
			label:'After fishing a crystal, edit its notes and crystallographic information',
			helpText:'If this is turned off, Icebear will add a note describing the fishing action.',
			value:Fishing.config.shownotesboxafterfish
		})
		
		f.select("label img").each(function(i){
			i.observe("click",function(){ 
				var inp=i.up("label").down("input");
				inp.id=inp.name;
				if(""==inp.value || isNaN(inp.value)){ inp.value=0; }
				Fishing.config[inp.name]=inp.value;
				Fishing[inp.name]=inp.value;
				inp.up("label").addClassName("updating");
				window.setTimeout(Fishing.saveConfig,50);
			});
		})
	},

	/***************************************************************************
	 * 
	 * SCANNING BARCODES AND RETRIEVING/RENDERING OBJECTS
	 * 
	 ***************************************************************************/
	
	/**
	 * Initiates the process of finding and rendering scanned items
	 */
	handleScan(){
		var b=$("barcode");
		var barcode=b.value.trim();
		if(""==barcode){ return false; }
		var conflict=false;
		Fishing.containersOnBench.each(function(c){
			if(c==barcode){
				alert(barcode+" is already on the bench, perhaps in another container?");
				b.value="";
				conflict=true;
			}
		});
		if(conflict){ return false; }
		new Ajax.Request('/api/plate/name/'+barcode,{
			method:"get",
			onSuccess:function(transport){
				if(!transport.responseJSON || !transport.responseJSON.rows){
					return AjaxUtils.checkResponse(transport);
				}
				Fishing.renderPlate(transport.responseJSON.rows[0]);
			},
			onFailure:function(){ 
				new Ajax.Request('/api/container/name/'+barcode,{
					method:"get",
					onSuccess:function(transport){
						if(!transport.responseJSON || !transport.responseJSON.rows){
							return AjaxUtils.checkResponse(transport);
						}
						var container=transport.responseJSON.rows[0];
						Fishing.renderContainer(container);
					},
					onFailure:function(){
						alert("No plate or container with barcode "+barcode);
					}
				});
			}
		});
		b.value="";
	},
	
	renderPlate:function(plate){
		$("platebox").down("h2").addClassName("updating");
		Fishing.plate=plate;
		Fishing.plateDropsRetrieved=false;
		Fishing.plateScoresRetrieved=false;
		Fishing.plateCrystalsRetrieved=false;
		new Ajax.Request('/api/plate/'+plate.id+'/welldrop',{
			method:"get",
			onSuccess:function(transport){
				Fishing.plate.welldrops=transport.responseJSON.rows;
				Fishing.plateDropsRetrieved=true;
			}
		});
		new Ajax.Request('/api/plate/'+plate.id+'/crystal?pagenum=1&pagesize=10000',{
			method:'get',
			onSuccess:function(transport){
				Fishing.plate.crystals=transport.responseJSON.rows,
				Fishing.plateCrystalsRetrieved=true;
			},
			onFailure:function(transport){
				Fishing.plate.crystals=[],
				Fishing.plateCrystalsRetrieved=true;
			}
		});
		new Ajax.Request('/api/plate/'+plate.id+'/dropbestscore?pagenum=1&pagesize=10000',{
			method:'get',
			onSuccess:function(transport){
				Fishing.plate.scores=transport.responseJSON.rows,
				Fishing.plateScoresRetrieved=true;
			},
			onFailure:function(transport){
				Fishing.plate.scores=[],
				Fishing.plateScoresRetrieved=true;
			}
		});
		window.setTimeout(Fishing.doRenderPlate,100);
	},
	doRenderPlate:function(){
		var plate=Fishing.plate;
		if(!plate){ return false; }
		if(!Fishing.plateDropsRetrieved || !Fishing.plateScoresRetrieved || !Fishing.plateCrystalsRetrieved){ 
			window.setTimeout(Fishing.doRenderPlate,100);
			return false; 
		}
		var box=$("platebox");
		if(!Fishing.plate || !Fishing.plate.welldrops){
			window.setTimeout(Fishing.renderPlate, 50);
			return false;
		}
		var header=box.down("h2");
		var boxbody=box.down(".boxbody");
		
		//whole boxbody goes up under header when platewidget is called. Correct for it.
		boxbody.absolutize();
		boxbody.style.top=header.getHeight()+"px";
		
		boxbody.innerHTML="";
		var allWellsHaveConstructs=true;
		Fishing.plate.welldrops.each(function(wd){
			if(""==wd.constructid){
				allWellsHaveConstructs=false;
			}
		});
		if(!allWellsHaveConstructs){
			boxbody.innerHTML+='<div class="warning">Plate has no protein/construct. This must be set before crystals are shipped.</div>';
		}
		var pw=PlateWidget.render({
			plateType:Fishing.plate,
			wellDrops:Fishing.plate.welldrops,
			dropPickerPosition:"left"
		}, boxbody);
		pw.plate=pw.plateType;
		header.innerHTML="Plate: "+plate.name;
		var drops=pw.down(".pw_plate").select(".pw_drop");
		drops.each(function(dropDiv){
			dropDiv.id="welldrop"+dropDiv.dataset.welldropid; //for direct targeting later
			dropDiv.crystals=[];
		});
		Fishing.plate.crystals.each(function(c){
			$("welldrop"+c.welldropid).crystals.push(c);
		});
		Fishing.plate.scores.each(function(s){
			$("welldrop"+s.id).score=s;
		});
		drops.each(function(dropDiv){
			Fishing.renderPlateWidgetDrop(dropDiv);
		});
		header.removeClassName("updating");
		Fishing.mapDragDropElements();
	},
	
	renderPlateWidgetDrop:function(dropDiv){
		dropDiv.innerHTML="";
		dropDiv.id="welldrop"+dropDiv.dataset.welldropid; //for direct targeting later
		dropDiv.onclick=function(evt){ Fishing.showImageAndCrosshairsForDropDiv(evt.target); };
		var xtalcount=document.createElement("div");
		xtalcount.addClassName("fish_xtalcount empty");
		var s=document.createElement("span");
		xtalcount.appendChild(s);
		dropDiv.appendChild(xtalcount);
		var xtaldrag=document.createElement("div");
		xtaldrag.addClassName("fish_xtaldrag");
		xtaldrag.dataset.welldropid=dropDiv.dataset.welldropid;
		dropDiv.appendChild(xtaldrag);
		if(dropDiv.score){
			dropDiv.style.backgroundColor="#"+dropDiv.score.bestscorecolor;
		}
		var numCrystals=dropDiv.crystals.length;
		if(numCrystals>0){
			xtalcount.down("span").innerHTML=numCrystals;
			xtalcount.removeClassName("empty");
		}
	},
	
	/**
	 * Initiates rendering of a pin, puck or dewar, fetching and attaching its contents.
	 * @param container A JSON object as returned by GET /api/container/1234.
	 * @param bypassDupeCheck If set, don't check whether this container is already on the bench (e.g., after nesting existing containers)
	 */
	renderContainer:function(container, bypassDupeCheck){
		new Ajax.Request('/api/container/'+container.id+'/fullcontent',{
			method:'get',
			onSuccess:function(transport){
				var containers=transport.responseJSON.rows;
				var conflict=false;
				if(!bypassDupeCheck){
					containers.each(function(c){
						//TODO Recursively
						if(conflict || !c.name || !c.containertypename){ return; }
						if(-1!=Fishing.containersOnBench.indexOf(c.name) && 0!==c.name.indexOf("dummypin")){
							alert(container.name+" contains "+c.name+", which is already on the bench. Cannot add "+container.name);
							conflict=true;
							return;
						}
					});
				}
				if(conflict){ return false; }
				containers.each(function(c){
					//TODO Recursively
					if(!c.name || !c.containertypename){ return; }
					if(-1==Fishing.containersOnBench.indexOf(c.name)){
						Fishing.containersOnBench.push(c.name);
					}
				});
				if(-1==Fishing.containersOnBench.indexOf(container.name)){
					Fishing.containersOnBench.push(container.name);
				}
				Fishing.doRenderContainer(container, containers);
			},
			onFailure:function(transport){
				if(404!=transport.status){
					return AjaxUtils.checkResponse(transport);
				}
				Fishing.doRenderContainer(container, null);
			},
		});
	},
	
	/**
	 * Renders a pin, puck or dewar to the UI, with contents attached.
	 * @param container A JSON object as returned by GET /api/container/1234
	 * @param contents An array of the container's contents.
	 */
	doRenderContainer(container,contents){
		var elem;
		var category=container.containercategoryname.toLowerCase();
		if(!$("container"+container.id)){
			elem=$(document.createElement("div"));
			elem.id="container"+container.id;
			elem.addClassName("fish_container fish_"+category);
			elem.dataset.itemname=container.name;
			elem.innerHTML='<span class="fish_contentcount"><span></span></span><span>'+container.name+'</span>';
			var box=$(category+"sbox");
			box=box.down(".boxbody")||box;
			box.appendChild(elem);
		} else {
			elem=$("container"+container.id);
			elem.removeClassName("fish_activedroppable");
		}
		elem.container=container;
		var contentCount=0;
		var positions=1*container.positions;
		if(contents){
			contents.each(function(c){
				if(""!=c && "dummy"!=c){
					contentCount++;
				}
			});
		} else {
			contents=["dummy"];
			for(var i=0;i<positions;i++){
				contents.push("");
			}
		}
		elem.container.childitems=contents;
		var cc=elem.down(".fish_contentcount");
		if("pin"==category){
			if(0==contentCount){
				elem.removeClassName("hascrystal");
			} else {
				elem.addClassName("hascrystal");
			}
		} else if("puck"==category){
			cc.innerHTML=contentCount+"<span>/"+positions+"</span>";
		} else if("dewar"==category){
			cc.innerHTML=contentCount;
			Fishing.showShipDewarButton();
		}
		cc.onclick=function(evt){
			Container.showAllContents(elem.container);
		};
		if(contentCount==positions){
			elem.addClassName("fish_full");
			elem.removeClassName("fish_notfull");
		} else {
			elem.removeClassName("fish_full");
			elem.addClassName("fish_notfull");
		}
		window.setTimeout(Fishing.mapDragDropElements,50);
		return elem;
	},

	
	/***************************************************************************
	 * 
	 * SHOWING DROP IMAGE AND SHOWING/ADDING CROSSHAIRS
	 * 
	 ***************************************************************************/
	
	currentImageRow:null,
	currentImageCol:null,
	currentImageDrop:null,
	showImageAndCrosshairsForDropDiv:function(dropDiv){
		if(dropDiv.up(".pw_drop")){ dropDiv=dropDiv.up(".pw_drop"); }
		Fishing.showImageAndCrosshairs(dropDiv.dataset.row, dropDiv.dataset.col, dropDiv.dataset.dropnumber);
	},
	showImageAndCrosshairs:function(row,col,dropNum){
		Fishing.currentImageRow=row;
		Fishing.currentImageCol=col;
		Fishing.currentImageDrop=dropNum;
		currentDrop=$("platebox").down(".platewidget").setCurrentDrop(row,col,dropNum);
		if(currentDrop.dropImages){
			Fishing.renderImageAndCrosshairs(currentDrop);
		} else {
			Fishing.findAndAttachImages(currentDrop);
		}
	},
	findAndAttachImages:function(dropDiv){
		new Ajax.Request('/api/welldrop/'+dropDiv.dataset.welldropid+'/timecourseimage',{
			method:'get',
			onSuccess:function(transport){ Fishing.findAndAttachImages_onSuccess(transport,dropDiv) },
			onFailure:function(transport){ Fishing.findAndAttachImages_onFailure(transport,dropDiv) },
		});
	},
	findAndAttachImages_onSuccess:function(transport,dropDiv){
		dropDiv.dropImages=transport.responseJSON.rows;
		Fishing.renderImageAndCrosshairs(dropDiv);
	},
	findAndAttachImages_onFailure:function(transport,dropDiv){
		if(404!=transport.status){
			AjaxUtils.checkResponse(transport);
			return false;
		}
	},
	
	renderImageAndCrosshairs:function(dropDiv){
		var box=$("platebox");
		var boxbody=box.down(".boxbody");
		var header=box.down("h2");
		header.oldInnerHTML=header.innerHTML;
		var newHeader='<img src="/skins/default/images/icons/no.gif" class="closeicon" onclick="Fishing.closeImageAndCrosshairs()">'+header.oldInnerHTML+
				" well "+Plate.getWellName(dropDiv.dataset.row, dropDiv.dataset.col, dropDiv.dataset.dropnumber);
		header.innerHTML=newHeader;
		var numImages=dropDiv.dropImages.length;
		//Default to last image, then try to find last visible
		var dropImage=dropDiv.dropImages[numImages-1];
		for(var i=numImages;i>0;i--){
			if("visible"==dropDiv.dropImages[i-1].lighttype.toLowerCase()){
				dropImage=dropDiv.dropImages[i-1];
				break;
			}
		}
		var imgDiv=document.createElement("div");
		$(imgDiv).addClassName("fish_dropimage");
		var img=document.createElement("img");
		img.src=dropImage.fullimageurl;
		img.id="fish_currentimage";
		img.dropImage=dropImage;
		img.observe("click", Fishing.markCrystalOnImage);
		imgDiv.appendChild(img);
		boxbody.appendChild(imgDiv);
		boxbody.style.borderRadius=0;
		imgDiv.pixelScale=img.getHeight()/dropImage.pixelheight;
		dropDiv.crystals.each(function(xtal){
			Fishing.drawCrosshair(xtal);
		});
		Fishing.mapDragDropElements();
		window.onresize=function(){ 
			Fishing.closeImageAndCrosshairs();
			Fishing.renderImageAndCrosshairs(dropDiv) 
		};
		/*
		 * TODO 
		 * Events for kb nav to other drops
		 * Nav in box header
		 * ....
		 */
		
	},
	drawCrosshair:function(xtal){
		var imgDiv=$("platebox").down(".fish_dropimage");
		var scale=imgDiv.pixelScale;
		var xh=$(document.createElement("div")).addClassName("fish_crosshair");
		var xhh=$(document.createElement("div")).addClassName("fish_xhhori");
		var xhv=$(document.createElement("div")).addClassName("fish_xhvert");
		var xhn=$(document.createElement("div")).addClassName("fish_xhnum");
		xh.appendChild(xhh);
		xh.appendChild(xhv);
		xh.appendChild(xhn);
		imgDiv.appendChild(xh);
		xhn.innerHTML=xtal.numberindrop;
		xh.style.left=(((1*xtal.pixelx)*scale)-(xh.getWidth()/2))+"px";
		xh.style.top =(((1*xtal.pixely)*scale)-(xh.getWidth()/2))+"px";
		xh.crystal=xtal;
		xh.id="crystal"+xtal.id;
		if(xtal.isfished==1){
			xh.addClassName("fish_alreadyfished"); 
			xh.title="Crystal has already been fished";
		}
		var drag=$(document.createElement("div")).addClassName("fish_xtaldrag");
		drag.crystal=xtal;
		xh.appendChild(drag);
	},
	
	closeImageAndCrosshairs:function(){
		Fishing.currentImageRow=null;
		Fishing.currentImageCol=null;
		Fishing.currentImageDrop=null;
		var box=$("platebox");
		var boxbody=box.down(".boxbody");
		var header=box.down("h2");
		header.innerHTML=header.oldInnerHTML;
		boxbody.down(".fish_dropimage").remove();
		window.onresize=null;
	},

	/***************************************************************************
	 * 
	 * CREATING CRYSTALS
	 * 
	 ***************************************************************************/
	
	/**
	 * Creates a new crystal, marked at the coordinates clicked on the image.
	 * @param evt The click event
	 */
	markCrystalOnImage:function(evt){
		var img=evt.target;
		//Crystal number should be highest number in the drop, +1.
		var xtalnum=0;
		var currentImage=$("fish_currentimage");
		currentImage.up(".fish_dropimage").select(".fish_crosshair").each(function(c){
			xtalnum=Math.max(xtalnum,c.crystal.numberindrop);
		});
		xtalnum++; 
		//Coordinates
		var offset=currentImage.cumulativeOffset();
		var scale=currentImage.getHeight() / img.dropImage.pixelheight;
		var crystalX=parseInt((evt.clientX-offset.left)/scale);
		var crystalY=parseInt((evt.clientY-offset.top)/scale);
		$("platebox").down("h2").addClassName("updating");
		var pw=$("platebox").down(".platewidget");
		var dropDiv=pw.down("tr.row"+pw.currentRow).down("td.col"+pw.currentCol).down("div.drop"+pw.currentDrop);
		Fishing.createCrystal({
			isDummyCoordinate:false,
			isFished:false,
			crystalX:crystalX,
			crystalY:crystalY,
			crystalNumber:xtalnum,
			projectId:currentImage.dropImage.projectid,
			wellDropId:currentImage.dropImage.welldropid,
			dropImageId:currentImage.dropImage.id
		});
	},
	
	createCrystal:function(crystal,container){
		new Ajax.Request('/api/crystal',{
			method:'post',
			parameters:{
				csrfToken:csrfToken,
				allowcreatedummyinspection:'yes',
				pixelx:crystal.crystalX,
				pixely:crystal.crystalY,
				isdummycoordinate:1*crystal.isDummyCoordinate,
				isfished:1*crystal.isFished,
				numberindrop:crystal.crystalNumber,
				projectid:crystal.projectId,
				welldropid:crystal.wellDropId,
				dropimageid:crystal.dropImageId,
				prefix:"dummy", //correct name generated server-side
				name:"dummy", //correct name generated server-side
			},
			onSuccess:function(transport){ Fishing.createCrystal_onSuccess(transport,container); },
			onFailure:function(transport){ Fishing.createCrystal_onFailure(transport,crystal,container); },
		});
	},
	createCrystal_onSuccess:function(transport,container){
		$("platebox").down("h2").removeClassName("updating");
		var crystal=transport.responseJSON.created;
		if(container){
			Fishing.addItemToContainer(crystal,container);
		}
		var drop=$("welldrop"+crystal.welldropid);
		if(drop){
			drop.crystals.push(crystal);
			Fishing.renderPlateWidgetDrop(drop);
			if($("fish_currentimage")){
				Fishing.closeImageAndCrosshairs();
				Fishing.showImageAndCrosshairsForDropDiv(drop);
			}
		}
		Fishing.mapDragDropElements();
	},
	createCrystal_onFailure:function(transport,crystal,container){
		$("platebox").down("h2").removeClassName("updating");
		AjaxUtils.checkResponse(transport);
		PlateWidget.renderPlateWidgetDrop($("welldrop"+crystal.wellDropId));
	},


	/***************************************************************************
	 * 
	 * DRAG-DROP FUNCTIONALITY
	 * 
	 ***************************************************************************/
	
	/**
	 * Defines which elements can be dragged onto which. For example, crystals into 
	 * pins or pucks but not dewars or other crystals; pucks into dewars only.
	 */
	mapDragDropElements:function(){
		window.setTimeout(function(){
			Fishing.makeDragImageDiv();
			var crystalDraggables=$$(".fish_xtaldrag");
			var crystalDroppables=$$(".fish_pin.fish_notfull, .fish_puck.fish_notfull");
			var pinDraggables=$$(".fish_pin");
			var pinDroppables=$$(".fish_puck.fish_notfull");
			var puckDraggables=$$(".fish_puck");
			var puckDroppables=$$(".fish_dewar.fish_notfull");
			Fishing.attachDroppables(crystalDraggables, crystalDroppables);
			Fishing.attachDroppables(pinDraggables, pinDroppables);
			Fishing.attachDroppables(puckDraggables, puckDroppables);
			crystalDraggables.each(function(d){
				d.ondragstart=Fishing.dragCrystalStart;
				d.ondragend=Fishing.dragCrystalEnd;
			});
			crystalDroppables.each(function(d){
				d.ondragover=Fishing.dropContainerOver;
				d.ondragenter=Fishing.dropContainerEnter;
				d.ondragleave=Fishing.dropContainerLeave;
				d.ondrop=Fishing.dropOntoContainer;
			});
			pinDraggables.each(function(d){ 
				d.ondragstart=Fishing.dragContainerStart; 
				d.ondragend=Fishing.dragContainerEnd; 
			});
			puckDraggables.each(function(d){
				d.ondragstart=Fishing.dragContainerStart; 
				d.ondragend=Fishing.dragContainerEnd; 
			});
			pinDroppables.each(function(d){
				d.ondragover=Fishing.dropContainerOver;
				d.ondragenter=Fishing.dropContainerEnter;
				d.ondragleave=Fishing.dropContainerLeave;
				d.ondrop=Fishing.dropOntoContainer;
			});
			puckDroppables.each(function(d){
				d.ondragover=Fishing.dropContainerOver;
				d.ondragenter=Fishing.dropContainerEnter;
				d.ondragleave=Fishing.dropContainerLeave;
				d.ondrop=Fishing.dropOntoContainer;
			});
		},50);
	},

	/**
	 * Puts a hidden div in the document, for use as a "drag image". Its background-image
	 * will be shifted by CSS classes to indicate what is being dragged.
	 */
	makeDragImageDiv:function(){
		if($("fish_dragimage")){
			return false;
		}
		var d=document.createElement("div");
		d.id="fish_dragimage";
		$("platebox").down(".boxbody").insert({ before:d });
		
	},
	
	/**
	 * Attaches an array of the suitable drop targets to each of the draggables.
	 */
	attachDroppables:function(draggables, droppables){
		if(!draggables || !droppables){ return false; }
		draggables.each(function(drag){
			drag.draggable=true;
			drag.droppables=droppables;
		});
	},
	
	dragCrystalStart:function(evt){
		var dragged=evt.target;
		if(dragged.up(".fish_alreadyfished")){
			alert("This crystal has already been fished.");
			return false;
		}
		Fishing.currentFish={};
		dragged.addClassName("fish_draggingcrystal");
		evt.dataTransfer.dropEffect="move";
		evt.dataTransfer.setData("text/plain", dragged.up(".fish_crosshair,.pw_drop").id);
		var dragIcon=$("fish_dragimage");
		dragIcon.addClassName("dragging fish_crystal");
		evt.dataTransfer.setDragImage(dragIcon, -10, 60);
		dragged.droppables.each(function(d){ d.addClassName("fish_droppable"); });
		if(dragged.crystal){
			Fishing.currentFish.crystal=dragged.crystal;
			Fishing.currentFish.wellDrop=$("welldrop"+dragged.crystal.welldropid).wellDrop;
		} else if(dragged.wellDrop) {
			Fishing.currentFish.wellDrop=dragged.wellDrop;
		}
	},
	dragCrystalEnd:function(evt){
		var dragIcon=$("fish_dragimage");
		dragIcon.removeClassName("fish_pin");
		dragIcon.removeClassName("fish_puck");
		dragIcon.removeClassName("fish_dewar");
		dragIcon.removeClassName("fish_crystal");
		$$(".fish_droppable").each(function(d){ d.removeClassName("fish_droppable"); });
		return false;
	},
	dragContainerStart:function(evt){
		var dragged=evt.target;
		if(dragged.hasClassName("fish_dropped")){ return false; }
		Fishing.currentFish={};
		evt.dataTransfer.dropEffect="move";
		evt.dataTransfer.setData("text/plain", dragged.id);
		dragged.addClassName("fish_draggingcontainer");
		var dragIcon=$("fish_dragimage");
		dragIcon.addClassName("fish_dragging fish_"+dragged.container.containercategoryname.toLowerCase());
		dragged.droppables.each(function(d){ d.addClassName("fish_droppable"); });
		evt.dataTransfer.setDragImage(dragIcon, -10, 60);
	},
	dragContainerEnd:function(evt){
		var dragged=evt.target;
		if(!dragged.hasClassName("fish_dropped")){
			dragged.removeClassName("fish_draggingcontainer");
		}
		var dragIcon=$("fish_dragimage");
		dragIcon.removeClassName("fish_pin");
		dragIcon.removeClassName("fish_puck");
		dragIcon.removeClassName("fish_dewar");
		dragIcon.removeClassName("fish_crystal");
		dragIcon.removeClassName("fish_dragging");
		$$(".fish_droppable").each(function(d){ d.removeClassName("fish_droppable"); });
		return false;
	},
	dropContainerOver:function(evt){
		if(evt.preventDefault){ evt.preventDefault(); }
		var droppable=$(evt.target);
		if(!droppable || !droppable.up){ return false; }
		if(droppable.up(".fish_container")){ droppable=droppable.up(".fish_container"); }
		if(droppable.hasClassName("fish_droppable")){
			evt.dataTransfer.dropEffect="move";
		} else {
			evt.dataTransfer.dropEffect="none";
		}
		return false;
	},
	dropContainerEnter:function(evt){
		var droppable=$(evt.target);
		if(!droppable || !droppable.up){ return false; }
		if(droppable.up(".fish_container")){ droppable=droppable.up(".fish_container"); }
		if(droppable.hasClassName("fish_droppable")){
			droppable.addClassName("fish_activedroppable");
		}
		return false;
	},
	dropContainerLeave:function(evt){
		var droppable=$(evt.target);
		if(!droppable || !droppable.up){ return false; }
		if(!droppable){ return false; }
		droppable.removeClassName("fish_activedroppable");
		return false;
	},
	
	/**
	 * Handle a draggable item being dropped onto its destination.
	 */
	dropOntoContainer:function(evt){
		evt.preventDefault();
		var destination=$(evt.target);
		var dropped=$(evt.dataTransfer.getData("text"));
		if(!destination || !destination.container || !dropped){ return false; };
		destination.removeClassName("fish_updated fish_activedroppable");
		var container=destination.container;
		var contained=null;
		if(dropped.container){
			contained=dropped.container;
		} else if(dropped.crystal){
			contained=dropped.crystal;
		} else if(dropped.crystals){
			contained=dropped.wellDrop;
		}
		if(!contained){ return false; }
		Fishing.addItemToContainer(contained, container);
		return false;
	},

	/***************************************************************************
	 * 
	 * ADDING ITEMS TO CONTAINERS
	 * 
	 ***************************************************************************/

	/**
	 * Returns the number of the first empty position, or -1 if the container is full.
	 * @param container An object representing the container, including a childitems array.
	 * @return The number of the first empty position, or -1 if full. This is 1-based, i.e., 7 is position 7 in the container.
	 */
	getFirstEmptyPosition:function(container){
		if(!container.childitems){
			if(container.container){
				//Maybe got the element, not the object
				container=container.container;
			}
		}
		if(!container || !container.childitems){
			alert("No childitems on container in getFirstEmptyPosition"); 
			return false;
		}
		var firstEmpty=-1;
		var slots=container.childitems;
		for(i=0;i<slots.length;i++){
			var slot=slots[i];
			if(""==slot || slot.isEmpty){
				slot="";
				firstEmpty=i;
				break;
			}
		}
		if(0<firstEmpty){
			Fishing.currentFish[container.containercategoryname.toLowerCase()+"position"]=firstEmpty;
		}
		return firstEmpty;
	},


	/**
	 * Adds the contained item into the first empty position of the containing item.
	 * Depending on configuration settings, the containing item may itself be moved
	 * into another container, e.g., crystal to pin to puck.
	 * @param contained Object representing the child item
	 * @param container Object representing the parent item
	 */
	addItemToContainer:function(contained, container){
		var slot=Fishing.getFirstEmptyPosition(container);
		if(-1==slot){
			alert("Container is full.");
			//drag-drop setup should prevent. No other handling needed.
			return false;
		}
		new Ajax.Request('/api/containercontent',{
			method:'post',
			parameters:{
				csrfToken:csrfToken,
				parent:container.id,
				child:contained.id,
				position:slot
			}, 
			onSuccess:function(transport){ Fishing.addItemToContainer_onSuccess(transport, contained, container); },
			onFailure:Fishing.addItemToContainer_onFailure,
		});
	},
	
	addItemToContainer_onSuccess:function(transport, contained, container){
		if($("container"+contained.id)){
			//pin or puck was dragged
			$("container"+contained.id).remove();
		} else if($("crystal"+contained.id)){
			//crystal crosshair was dragged
			$("crystal"+contained.id).crystal.isfished=1;
			$("crystal"+contained.id).addClassName("fish_alreadyfished");
			Fishing.renderPlateWidgetDrop($("welldrop"+contained.welldropid));
		} else if($("welldrop"+contained.id)){
			//well drop was dragged, should have created crystal
			var created=transport.responseJSON.created;
			created.each(function(obj){
				if("crystal"==obj.objecttype){
					$("welldrop"+contained.id).crystals.push(obj);
					Fishing.currentFish.crystal=obj;
				} else if("container"==obj.objecttype && "pin"==obj.containercategoryname.toLowerCase()){
					Fishing.currentFish.pin=obj;
				}
			});
			Fishing.renderPlateWidgetDrop($("welldrop"+contained.id));
		}
		Fishing.currentFish[container.containercategoryname.toLowerCase()]=container;
		$("container"+container.id).addClassName("fish_updated");
		window.setTimeout(function(){
			Fishing.renderContainer(container, true);
		}),50;
		if(1==Fishing.config.autopintopuck && "pin"==container.containercategoryname.toLowerCase()){
			var firstPuckWithSpace=$("pucksbox").down(".fish_notfull");
			if(firstPuckWithSpace){
				window.setTimeout(function(){
					Fishing.addItemToContainer(container, firstPuckWithSpace.container);
				},300);
			}
		} else if(Fishing.currentFish.crystal){
			Fishing.addDefaultNote();
			if(1==Fishing.config.shownotesboxafterfish){
				Fishing.showNotesBox();
			}
		}
	},
	addItemToContainer_onFailure:function(transport){
		AjaxUtils.checkResponse(transport);
	},

	/***************************************************************************
	 * 
	 * CRYSTAL NOTES AND DETAILS
	 * 
	 ***************************************************************************/

	saveNoteOnLastFishedCrystal:function(noteText, successCallback){
		var crystal=Fishing.currentFish.crystal;
		if(""==noteText){ return false; }
		new Ajax.Request('/api/note',{
			method:'post',
			parameters:{
				csrfToken:csrfToken,
				parentid:crystal.id,
				projectid:crystal.projectid,
				text:noteText
			},
			onSuccess:function(transport){ if(successCallback){ successCallback(transport.responseJSON); } },
			onFailure:AjaxUtils.checkResponse
		})
	},

	addDefaultNote:function(){
		if(!Fishing.currentFish || !Fishing.currentFish.crystal){ return false; }
		if(!Fishing.currentFish.wellDrop){
			Fishing.currentFish.wellDrop=$("welldrop"+Fishing.currentFish.crystal.welldropid).wellDrop;
		}
		var defaultNote='Crystal fished from plate '+Fishing.plate.name+' well '+Plate.getWellName(Fishing.currentFish.wellDrop.row, Fishing.currentFish.wellDrop.col, Fishing.currentFish.wellDrop.dropnumber);
		if(Fishing.currentFish.pin && 0!==Fishing.currentFish.pin.name.indexOf("dummy")){
			defaultNote+=' onto pin '+Fishing.currentFish.pin.name;
		}
		if(Fishing.currentFish.puck){
			defaultNote+=' and put into position '+Fishing.currentFish.puckposition+' of puck '+Fishing.currentFish.puck.name;
		}
		defaultNote+='.';
		Fishing.saveNoteOnLastFishedCrystal(defaultNote);
	},

	autoNoteSaveDelay:10, //seconds
	
	doNoteCountdown:function(){
		if(!$("autonotesavetext")){ return false; }
		var secondsLeft=parseInt($("autonotesavedelay").innerHTML);
		if(0==secondsLeft){
			ui.closeModalBox();
		} else {
			$("autonotesavedelay").innerHTML=secondsLeft-1;
			window.setTimeout(Fishing.doNoteCountdown,1000);
		}
	},
	
	cancelAutoNoteSave:function(){
		if($("autonotesavetext")){ $("autonotesavetext").remove(); }
	},

	
	showNotesBox:function(){
		Fishing.showingNotes=true;
		var mb=ui.modalBox({
			title:'Fishing complete',
			content:'',
			onclose:Fishing.hideNotesBox
		});
		var cs=getComputedStyle($("pucksdewarsbox"));
		$("modalContent").style.top=0;
		$("modalContent").style.left=0;
		$("modalContent").style.right=0;
		$("modalContent").style.bottom=0;
		$("modalContent").style.height="100%";
		$("modalContent").style.width="100%";
		$("modalBox").style.left=$("pinsbox").cumulativeOffset().left+"px";
		$("modalBox").style.top=$("pinsbox").cumulativeOffset().top+"px";

		var frm=mb.form({
			action:'#',
			method:'post',
			autosubmit:false
		});
		
		frm.onsubmit=function(){ ui.closeModalBox(); return false; }
		
		var ta=frm.textArea({
			label:'Crystal treatment',
			name:'fishingnote',
			value:'',
			helpText:'Describe the cryoprotection, soaking, etc. This text is not sent to the synchrotron.'
		});
		ta.style.textAlign="center";
		ta.down("textarea").style.width="95%";
		ta.down("textarea").style.height="6em";
		ta.down("textarea").observe("focus", Fishing.cancelAutoNoteSave);
		ta.observe("click", Fishing.cancelAutoNoteSave);

		Crystal.writeDetailsFormFields(Fishing.currentFish.crystal, true, frm);

		var sub=frm.formField({
			label:'<span id="autonotesavetext">This box will close automatically in <span id="autonotesavedelay">'+(1+Fishing.autoNoteSaveDelay)+'...</span>',
			content:'<input type="submit" id="savenotes" name="savenotes" value="Save notes" onclick="Fishing.saveClicked=true" />',
		});
		ui.addSuppliedHelpText(sub, 'Editing anything in this box cancels the timer.');
		Fishing.saveClicked=false;
		
		window.setTimeout(function(){
			frm.select("button,input,textarea,select").each(function(inp){
				inp.observe("click", Fishing.cancelAutoNoteSave);
				inp.observe("focus", Fishing.cancelAutoNoteSave);
			});
			Fishing.doNoteCountdown();
		},10);
	},
	
	/**
	 * Callback for ui.closeModalBox.
	 */
	hideNotesBox:function(){
		var noteText=$("fishingnote").value.trim();
		if(""==noteText){
			Fishing.showingNotes=false;
			return true;
		}
		if(Fishing.saveClicked){
			$("savenotes").up("label").addClassName("updating");
			Fishing.saveNoteOnLastFishedCrystal(noteText);
			Fishing.showingNotes=false;
			return true;
		}
		if(!confirm("Your notes have not been saved.\n\nClose without saving?")){
			return false; //leave the box open
		}
		Fishing.showingNotes=false;
		return true;
	},

	/***************************************************************************
	 * 
	 * ONWARD SHIPPING
	 * 
	 ***************************************************************************/

	/**
	 * Shows a "Ship this|these dewar(s)" link if dewars are present and user can ship.
	 */
	showShipDewarButton:function(){
		/**
		 * Shipping functionality hidden for release, so...
		 */
		return false;
		
		var dewars=$$(".fish_dewar");
		if(!dewars || !canShip){
			if($("fish_shipdewars")){
				$("fish_shipdewars").remove();
			}
		} else {
			if(!$("fish_shipdewars")){
				var sd=document.createElement("label");
				sd.id="fish_shipdewars";
				var a=document.createElement("a");
				a.innerHTML="Ship this dewar...";
				sd.appendChild(a);
				$("dewarsbox").appendChild(sd);
			}
			if(1!=dewars.length){
				$("fish_shipdewars").down("a").innerHTML="Ship these dewars...";
			}
			var sdlink=$("fish_shipdewars").remove();
			$("dewarsbox").insert({"bottom":sdlink });
			$("fish_shipdewars").down("a").onclick=Fishing.shipDewars;
		}
	},

	
	/**
	 * Passes the names of the dewars to the shipment create UI.
	 */
	shipDewars:function(){
		if(!confirm("Leave this page and go to the shipping page?")){ return false; }
		var dewars=$$(".fish_dewar");
		var dewarNames=new Array();
		dewars.each(function(d){
			dewarNames.push(encodeURIComponent(d.container.name));
		});
		document.location.href="/shipment/create?dewars="+dewarNames.join(",");
	},

	
}